from math import ceil, floor
import numpy as np

from .mesh_toolkit import MeshToolKit

class DEMToolKit(MeshToolKit):

   class Points_out_of_DEM_range_error(Exception): pass

   def __init__(self, info):
      super().__init__( info=info)


   def load_DEM_file(self, DEM_file_name):
      self.info.start("Loading elevation grid file: " + DEM_file_name)
      self.file_name = DEM_file_name
      with open(self.file_name) as f:
            self.nx = int(f.readline().split()[1])
            self.ny = int(f.readline().split()[1])
            self.x0 = float(f.readline().split()[1])
            self.y0 = float(f.readline().split()[1])
            self.a = float(f.readline().split()[1])
            self.x_max = self.x0 + (self.nx - 1) * self.a
            self.y_max = self.y0 + (self.ny - 1) * self.a
            self.x_range = self.x_max - self.x0
            self.y_range = self.y_max - self.y0
            self.novalue = float(f.readline().split()[1])
      self.data = np.flip(np.loadtxt(self.file_name, dtype=np.float64, skiprows=6), axis=0).transpose()
      self.info("x0 : {:g}".format(self.x0))
      self.info("y0 : {:g}".format(self.y0))
      self.info("Columns count : {:n}".format(self.nx))
      self.info("Rows count : {:n}".format(self.ny))
      self.info("Cell size: {:g}".format(self.a))
      self.info("Min. value: {:g}".format(self.data.min()))
      self.info("Max. value: {:g}".format(self.data.max()))
      self.info("No values count: {:n}".format(np.sum(self.data==self.novalue)))
      self.info.OK()  
      

   def get_xyz_points(self, x0=None, y0=None, x_max=None, y_max=None):      
      self.info.start("Getting coordinates of DEM positions")
      if x0 is None: x0 = self.x0
      if y0 is None: y0 = self.y0
      if x_max is None: x_max = self.x_max
      if y_max is None: y_max = self.y_max
      a = self.a
      x_i0 = ceil((x0 - self.x0) / a)
      x_i1 = floor((x_max - self.x0) / a) + 1
      y_i0 = ceil((y0 - self.y0) / a)
      y_i1 = floor((y_max - self.y0) / a) + 1
      nx, ny = x_i1 - x_i0, y_i1 - y_i0
      p = np.empty((nx*ny, 3), np.float64)
      yy = np.arange(y_i0, y_i1, dtype=np.float64) * a + self.y0
      for x_i in range(x_i0, x_i1):
         s = slice((x_i - x_i0)*ny, (x_i-x_i0+1)*ny)
         p[s,0] = x_i * a + self.x0
         p[s,1] = yy
         p[s,2] = self.data[x_i, y_i0:y_i1]
      self.info.OK()
      return p


   def _validate_points_in_grid(self, p):
      n = np.sum(p[(p[:,0] < self.x0) | (p[:,0] > self.x_max) | (p[:,1] < self.y0) | (p[:,1] > self.y_max)])
      if n > 0:
         raise self.Points_out_of_DEM_range_error("Coordinates of {:n} points are outof range.".format(n))


   def get_points_elevations(self, points):
      self.info.start("Calculating elevations of {:n} points".format(len(points)))
      if "data" not in self.__dict__:
         raise self.ToolKitException("No DEM file is loaded.") 
      self._validate_points_in_grid(points)
      x0, y0, a, n = self.x0, self.y0, self.a, points.shape[0]
      e = self.data
      Q = np.empty(points.shape, np.int32)
      d, Q[:] = np.modf((points - [x0,y0]) / a)
      Qx, Qy = Q[:,0], Q[:,1]
      Cz = (e[Qx,Qy] + e[Qx+1,Qy] + e[Qx,Qy+1] + e[Qx+1,Qy+1]) / 4
      x, y = points[:,0] - x0, points[:,1] - y0
      z = np.empty(n, np.float64)
      k1, k2 = d[:,0] < d[:,1], d[:,0] > - d[:,1] + 1
      b = np.logical_not(k1) & np.logical_not(k2)
      z1, z2 = e[Qx[b], Qy[b]], e[Qx[b]+1, Qy[b]]
      z[b] = (z2 - z1) * (x[b]/a - Qx[b]) + (2*Cz[b]-z2-z1) * (y[b]/a - Qy[b]) + z1
      b = np.logical_not(k1) & k2
      z1, z2 = e[Qx[b]+1, Qy[b]], e[Qx[b]+1, Qy[b]+1]
      z[b] = (z1+z2-2*Cz[b]) * (x[b]/a - Qx[b] - 1) + (z2 - z1) * (y[b]/a - Qy[b]) + z1
      b = k1 & k2
      z1, z2 = e[Qx[b], Qy[b]+1], e[Qx[b]+1, Qy[b]+1]
      z[b] = (z2-z1) * (x[b]/a - Qx[b]) + (z2+z1-2*Cz[b]) * (y[b]/a - Qy[b] - 1) + z1
      b = k1 & np.logical_not(k2)
      z1, z2 = e[Qx[b], Qy[b]], e[Qx[b], Qy[b]+1]
      z[b] = (2*Cz[b]-z1-z2) * (x[b]/a - Qx[b]) + (z2 - z1) * (y[b]/a - Qy[b]) + z1
      self.info.OK()
      return z
      
      







