import numpy as np
from time import sleep

from .mesh import MeshForToolKits
from .params import ParamsToolKit
from .DEM import DEMToolKit
from .bands import BandsToolKit
from .edit import MeshEditToolKit
from .report import ReportToolKit
from .data import DataToolKit

class DiskmodTools:

   class DiskmodException(Exception): pass

   def __init__(self, info):
      self.info = info
      self.DEM = DEMToolKit(info)
      self.mesh = m = MeshForToolKits(info)
      self.params = ParamsToolKit(m)
      self.bands = BandsToolKit(self.params)
      self.mesh_edit = MeshEditToolKit(self.DEM, self.params, self.bands)      
      self.data = DataToolKit(self.params, self.DEM, self.bands)
      self.report = ReportToolKit(self.params, self.DEM, self.bands, self.data)


   # *********** loading data ********************

   def is_mesh_loaded(self):
      return "nodes_id" in self.mesh.__dict__

   def load_mesh(self, mesh_file, physicals_names_file):
      self.mesh.load_msh(mesh_file, physicals_names_file)
      self.mesh.update_above_toolkits()


   def is_DEM_loaded(self):
      return "data" in self.DEM.__dict__

   def load_DEM(self, file_name):
      self.DEM.load_DEM_file(file_name)
      self.DEM.update_above_toolkits()


   def is_top_boudary_setting(self):
      return len(self.params.OPTIONS_top_boundary_physicals_ids) > 0

   def get_top_boundary_physicals_selection(self):  
      return self.params.get_top_boundary_physicals_selection()
  
   def set_top_boundary_physicals_selection(self, physicals_ids):
      self.params.set_top_boundary_physicals_selection(physicals_ids)
      self.params.update_above_toolkits()

   
   def get_physicals_names_cross_sections(self):
      self.info.start("Getting 1D and 2D physicals names with cross sevtions")
      names = {p.id:[p.name, p.dim, self.params.OPTIONS_physicals_cross_sections.get(p.id)] for p in self.mesh.physicals.values() if p.dim < 3}
      self.info.OK()
      return names, self.params.OPTIONS_default_physicals_cross_section

   def set_physicals_names_cross_section(self, physicals_values, default):
      self.info.start("Setting crosssections of physicals")
      self.params.OPTIONS_physicals_cross_sections = physicals_values
      self.params.OPTIONS_default_physicals_cross_section = default
      self.info.OK()
  
   def get_reference_point(self):
      return self.params.x_ref, self.params.y_ref, self.params.x_0, self.params.y_0, self.params.x_max, self.params.y_max
  
   def set_reference_point(self, x, y):
      self.info.start("Setting reference position")
      self.params.x_ref, self.params.y_ref = x, y
      self.info.OK()
   

   # ************** mesh edit *************************

   def get_physicals_names(self):
      self.info.start("Getting physicals names")
      names = {p.id:(p.name, p.dim) for p in self.mesh.physicals.values()}
      self.info.OK()
      return names

   def report_model_range(self):
      r1 = self.report.report_model_range()
      r2 = self.report.report_DEM_range_of_mesh_area() if self.is_DEM_loaded() and self.is_top_boudary_setting() else (None, None)
      return r1 + r2
      
   def report_physicals_group(self, physicals_ids):
      return self.report.report_physicals_group(physicals_ids)

   def change_top_nodes_z_by_DEM(self, desc_file_name):
      self.mesh_edit.change_top_nodes_z_by_DEM(desc_file_name)
      self.mesh.update_above_toolkits()

   def change_nodes_z_from_level_by_DEM(self, z_level, desc_file_name):
      self.mesh_edit.change_nodes_z_from_level_by_DEM(z_level, desc_file_name)
      self.mesh.update_above_toolkits()
   
   def remove_physicals(self, physicals_ids):
      self.mesh_edit.remove_physicals(physicals_ids)
      self.mesh.update_above_toolkits()

   def merge_physicals(self, new_name, new_id, physicals_ids):
      self.mesh_edit.merge_physicals(new_name, new_id, physicals_ids)
      self.mesh.update_above_toolkits()

   def change_physical_id_name(self, id, new_id, new_name):
      self.mesh_edit.change_physical_id_name(id, new_id, new_name)
      self.mesh.update_above_toolkits()
     
   def save_mesh(self, file_name):
      self.mesh.save_msh(file_name)


   # *************** data ***********************

   def is_element_data(self):
      return len(self.data.e_data) > 0

   def get_element_data_references(self):
      return sorted(self.data.e_data)

   def get_element_data_info(self, data_refs=None):
      return self.data.get_e_data_info(data_refs=data_refs)

      
   def load_msh_element_data_references(self, file_name=None):
      return self.mesh.load_msh_element_data_references(file_name=file_name)

   def load_element_data_from_same_mesh(self, file_name, names=None, times=None, times_indices=None):
      return self.data.load_element_data_from_same_mesh(file_name, names=names, times=times, times_indices=times_indices)

   def load_element_data_from_any_mesh(self, file_name, names=None, times=None, times_indices=None):
      return self.data.load_element_data_from_any_mesh(file_name, names=names, times=times, times_indices=times_indices)

   def get_element_data_references_physicals_names(self):      
      self.info.start("Getting element data references physicals names")
      data_refs = sorted(self.data.e_data)
      physical_names = {p.id:(p.name, p.dim) for p in self.mesh.physicals.values()}
      self.info.OK()
      return data_refs, physical_names


   def save_partial_element_data(self, file_name, data_refs=None, physicals_ids=None, x_min=None, x_max=None, y_min=None, y_max=None, xy_by_ref=False, z_min=None, z_max=None, depth_min=None, depth_max=None, depth_by_DEM=False, center_in_range=True):
      self.info.start("Saving partial element data to file '" + file_name + "'")
      if xy_by_ref:
         x_min, y_min = self.params.get_xy_by_ref(x_min, y_min)
         x_max, y_max = self.params.get_xy_by_ref(x_max, y_max)      
      if data_refs is None: data_refs = sorted(self.data.e_data)         
      elems = self.report.select_elements(physicals_ids=physicals_ids, x_min=x_min, x_max=x_max, y_min=y_min, y_max=y_max, z_min=z_min, z_max=z_max, depth_min=depth_min, depth_max=depth_max, depth_by_DEM=depth_by_DEM, center_in_range=center_in_range)
      ext = file_name.split(".")[-1].upper()
      if ext == "MSH":
         self.mesh.save_msh_with_selected_elements(file_name, elems)      
         for r in data_refs:
            self.mesh.append_element_data_as_one_name(file_name, *r[1:], elems, self.data.e_data[r][elems])
      elif ext == "H5":
         self.report.save_data_to_h5(elems, data_refs, file_name)
      else: raise self.DiskmodException("Invalid file extension.")
      self.info.OK()      


   def save_partial_element_data_with_complete_mesh(self, file_name, data_refs=None, physicals_ids=None, x_min=None, x_max=None, y_min=None, y_max=None, xy_by_ref=False, z_min=None, z_max=None, depth_min=None, depth_max=None, depth_by_DEM=False, center_in_range=True):
      self.info.start("Saving partial element data with complete mesh to file '" + file_name + "'")
      if xy_by_ref:
         x_min, y_min = self.params.get_xy_by_ref(x_min, y_min)
         x_max, y_max = self.params.get_xy_by_ref(x_max, y_max)                         
      if data_refs is None: data_refs = sorted(self.data.e_data)
      elems = self.report.select_elements(physicals_ids=physicals_ids, x_min=x_min, x_max=x_max, y_min=y_min, y_max=y_max, z_min=z_min, z_max=z_max, depth_min=depth_min, depth_max=depth_max, depth_by_DEM=depth_by_DEM, center_in_range=center_in_range)      
      ext = file_name.split(".")[-1].upper()
      if ext == "MSH":
         self.mesh.save_msh(file_name)  
         for r in data_refs:
            self.mesh.append_element_data_as_one_name(file_name, *r[1:], elems, self.data.e_data[r][elems])
      elif ext == "H5":
         self.report.save_data_to_h5(elems, data_refs, file_name)
      else: raise self.DiskmodException("Invalid file extension.")
      self.info.OK()   


   def set_e_data_boundary_e2_physicals_as_e4_neighbours(self, data_refs, e2_boundary_physicals_ids):
      e2 = self.mesh.get_elements_of_physicals(e2_boundary_physicals_ids)
      r = self.data.set_e_data_boundary_e2_as_e4_neighbours(data_refs, e2)
      return r


   def get_vector_element_data_references(self):
      return [r for r, d in self.data.e_data.items() if d.ndim == 2]


   def create_scalar_e_data_from_3D_vectors(self, vector_e_data_refs=None, x_part=True, y_part=True, z_part=True, vert_angle=True, horiz_angle=True, size=True):
       if vector_e_data_refs is None:
          vector_e_data_refs = sorted((r for r, d in self.data.e_data.items() if d.ndim == 2))
       self.data.create_scalar_e_data_from_3D_vectors(vector_e_data_refs, x_part=x_part, y_part=y_part, z_part=z_part, vert_angle=vert_angle, horiz_angle=horiz_angle, size=size)


   def calc_new_e_data_by_formula(self, data_refs, new_name, formula):
      return self.data.calc_new_e_data(data_refs, new_name, formula)


   def calc_e_data_statistics(self, physicals_ids=None, x_min=None, x_max=None, y_min=None, y_max=None, xy_by_ref=False, z_min=None, z_max=None, depth_min=None, depth_max=None, center_in_range=True, depth_by_DEM=False,
                                   data_refs=None, weights_type=None, quantiles=None, hist_bins=None, hist_as_pdf=False,
                                   file_name_rows=None, file_name_cols=None, include_data_file_name=False, to_protocol=True, col_sep=",", dec_comma=False, num_format="g"):
      self.info.start("Calculating statistics of selected element data")
      R = self.report
      if xy_by_ref:
         x_min, y_min = self.params.get_xy_by_ref(x_min, y_min)
         x_max, y_max = self.params.get_xy_by_ref(x_max, y_max)
      e = R.select_elements(physicals_ids=physicals_ids, x_min=x_min, x_max=x_max, y_min=y_min, y_max=y_max, z_min=z_min, z_max=z_max, depth_min=depth_min, depth_max=depth_max, depth_by_DEM=depth_by_DEM, center_in_range=center_in_range)
      r = R.get_e_data_statistics(elements=e, weights_type=weights_type, data_refs=data_refs, quantiles=quantiles, hist_bins=hist_bins, hist_as_pdf=hist_as_pdf)
      if file_name_rows is not None: R.print_statistics_as_table(file_name_rows, r, data_as_columns=False, include_data_file_name=include_data_file_name, col_sep=col_sep, dec_comma=dec_comma, num_format=num_format)
      if file_name_cols is not None: R.print_statistics_as_table(file_name_cols, r, data_as_columns=True, include_data_file_name=include_data_file_name, col_sep=col_sep, dec_comma=dec_comma, num_format=num_format)
      p = R.print_statistics_as_protocol_results(r, num_format=num_format) if to_protocol else None
      self.info.OK()
      return p
      


   # *********** water flow raw **********************

   def is_water_flow_raw_loaded(self):
      return len(self.data.__dict__.get("raw_ep_ef_sp_sf", {})) > 0

   def load_water_flow_raw(self, file_name):
      return self.data.load_water_flow_raw(file_name)

   def calc_e4_flux_over_e2e4_neighbours(self, e4_physicals_ids, e2e4_neighbours_physicals_ids, T, vector=None):
      return self.report.calc_e4_flux_over_e2e4_neighbours(e4_physicals_ids, e2e4_neighbours_physicals_ids, T, vector)


   # ************** raster and points data *****************
   
   def save_z_level_raster_data(self, z_level, file_name, data_refs=None, x_min=None, x_max=None, y_min=None, y_max=None, xy_by_ref=False, cell_size=10, cell_center=False, nbs_by_nodes_num=[], IDW_exp=2, value_instead_of_nan=None, num_format="g"):
      self.info.start("Saving Z level raster file '{:s}'".format(file_name))
      if xy_by_ref:
         x_min, y_min = self.params.get_xy_by_ref(x_min, y_min)
         x_max, y_max = self.params.get_xy_by_ref(x_max, y_max)      
      if data_refs is None: data_refs = sorted(self.data.e_data, key=lambda r: (r[0], r[1], r[3]))
      data = self.report.calc_z_level_grid_data(z_level, x_min, x_max, y_min, y_max, cell_size, cell_center, data_refs, nbs_by_nodes_num, IDW_exp)
      if file_name.upper().endswith(".ASC"):
         if len(data_refs) > 1: raise self.DiskmodException("When ASC file is saved, onlz one data reference must be selected.")
         self.report.seve_ASC_file(file_name, x_min, y_min, cell_size, np.nan if value_instead_of_nan is None else value_instead_of_nan, data[0], num_format)
      elif file_name.upper().endswith(".MSH"):
         if cell_center: self.report.save_msh_elements_grid_data_file(file_name, x_min, y_min, cell_size, data_refs, data, num_format)
         else: self.report.save_msh_nodes_grid_data_file(file_name, x_min, y_min, cell_size, data_refs, data, num_format)         
      self.info.OK()


   def save_depth_raster_data(self, depth, file_name, depth_by_DEM=False, data_refs=None, x_min=None, x_max=None, y_min=None, y_max=None, xy_by_ref=False, cell_size=10, cell_center=False, nbs_by_nodes_num=[], IDW_exp=2, value_instead_of_nan=None, num_format="g"):
      self.info.start("Saving depth  raster file '{:s}'".format(file_name))
      if xy_by_ref:
         x_min, y_min = self.params.get_xy_by_ref(x_min, y_min)
         x_max, y_max = self.params.get_xy_by_ref(x_max, y_max)      
      if data_refs is None: data_refs = sorted(self.data.e_data, key=lambda r: (r[0], r[1], r[3]))
      data = self.report.calc_depth_grid_data(depth, depth_by_DEM, x_min, x_max, y_min, y_max, cell_size, cell_center, data_refs, nbs_by_nodes_num, IDW_exp)
      if file_name.upper().endswith(".ASC"):
         if len(data_refs) > 1: raise self.DiskmodException("When ASC file is saved, onlz one data reference must be selected.")
         self.report.seve_ASC_file(file_name, x_min, y_min, cell_size, np.nan if value_instead_of_nan is None else value_instead_of_nan, data[0], num_format)
      elif file_name.upper().endswith(".MSH"):
         if cell_center: self.report.save_msh_elements_grid_data_file(file_name, x_min, y_min, cell_size, data_refs, data, num_format)
         else: self.report.save_msh_nodes_grid_data_file(file_name, x_min, y_min, cell_size, data_refs, data, num_format)         
      self.info.OK()


   def save_points_data(self, points, points_names, file_name, xy_by_ref=False, data_refs=None, nbs_by_nodes_num=[], IDW_exp=2, ref_file_name_col=False, col_sep="\t", dec_comma=False, num_format="g", write_to_protocol=True):
      if xy_by_ref:
         points = [(*self.params.get_xy_by_ref(x, y), z) for x, y, z in points]
      if data_refs is None: data_refs = sorted(self.data.e_data, key=lambda r: (r[0], r[1], r[3]))      
      return self.report.save_points_data(points, points_names, data_refs, nbs_by_nodes_num, IDW_exp, file_name, ref_file_name_col=ref_file_name_col, col_sep=col_sep, dec_comma=dec_comma, num_format=num_format, write_to_protocol=write_to_protocol)


   def save_points_data_with_depth(self, points_xy, points_depth, points_names, file_name, depth_by_DEM=False, xy_by_ref=False, data_refs=None, nbs_by_nodes_num=[], IDW_exp=2, ref_file_name_col=False, col_sep="\t", dec_comma=False, num_format="g", write_to_protocol=True):
      if xy_by_ref:
         points_xy = [self.params.get_xy_by_ref(x, y) for x, y in points_xy]
      if data_refs is None: data_refs = sorted(self.data.e_data, key=lambda r: (r[0], r[1], r[3]))      
      return self.report.save_points_data_with_depth(points_xy, points_depth, points_names, data_refs, nbs_by_nodes_num, IDW_exp, file_name, depth_by_DEM=depth_by_DEM, ref_file_name_col=ref_file_name_col, col_sep=col_sep, dec_comma=dec_comma, num_format=num_format, write_to_protocol=write_to_protocol)





