import numpy as np

from .mesh_toolkit import MeshToolKit


class MeshBand_Err:
   class Points_outside_mesh(Exception): pass

class BandsToolKit(MeshToolKit):

   NO_E = -1

   def __init__(self, params):
      super().__init__(params=params)
      self.OPTION_buf_points_elems_CRIT = 100000
      self.OPTION_buf_points_trias_CRIT = 100000


   # ********* creating bands - PRIVATE *****************

   def update_method(self):
      pass


   # ********* surface by triangles ************

   def get_buf_points_triangles(self, points, triangles=None):
      self.info.start("Building points & triangles buffer")
      r, m = self.params, self.mesh
      st_p, st_t = [[] for _ in range(4)], [[] for _ in range(4)]      
      triangles = np.arange(0, len(r.tria_nodes), dtype=np.int32)
      buf = [(np.arange(0, len(points), dtype=np.int32), triangles, r.x_0, r.y_0, r.x_range, r.y_range)]
      BUF = []      
      g = [None for _ in range(4)]
      w = (-1, -1)
      while len(buf) > 0:         
         for s in st_p: s.clear()
         for s in st_t: s.clear()
         pts, trias, x0, y0, xr, yr = buf.pop(0)         
         if w != (xr, yr):
            w = (xr, yr)
            self.info("Ranges x: {:g}  y: {:g}".format(*w))
         xh, yh = xr/2, yr/2
         xs, ys = x0 + xh, y0 + yh
         st_i = 1*(points[pts,0] > xs) + 2*(points[pts,1] > ys)
         for p, i in zip(pts, st_i): st_p[i].append(p)
         n = m.nodes_xyz[r.tria_nodes[trias],:2]         
         D, U = n.min(axis=1), n.max(axis=1)
         dx, dy = D[:,0] < xs, D[:,1] < ys
         ux, uy = U[:,0] >= xs, U[:,1] >= ys         
         for i in np.where(dx & dy)[0]: st_t[0].append(trias[i])
         for i in np.where(ux & dy)[0]: st_t[1].append(trias[i])
         for i in np.where(dx & uy)[0]: st_t[2].append(trias[i])
         for i in np.where(ux & uy)[0]: st_t[3].append(trias[i])                     
         g[0] = xs - xh, ys - yh
         g[1] = xs     , ys - yh
         g[2] = xs - xh, ys
         g[3] = xs     , ys         
         for p, t, b in zip(st_p, st_t, g):
            c = len(p) * len(t)
            if c > self.OPTION_buf_points_trias_CRIT:
               buf.append((np.array(p, dtype=np.int32), np.array(t, dtype=np.int32), *b, xh, yh))
            elif c > 0:
               BUF.append((np.array(p, dtype=np.int32), np.array(t, dtype=np.int32))) 
      self.info.OK()
      return BUF


   def get_surf_z(self, points):
      self.info.start("Calculating mesh surface elevations of {:n} points".format(len(points)))
      if self.params.tria_nodes is None: self.params.create_tria_nodes() 
      ab, c = self.params.get_tria_xy_sides_ab_c()
      planes_abc, planes_d = self.params.get_tria_planes_abc_d()
      p_z = np.full(len(points), np.nan, dtype=np.float64) 
      buf = self.get_buf_points_triangles(points)
      self.info.start("Processing {:n} items of buffer".format(len(buf)))
      for ps, ts in buf:
         ps_xy1 = np.empty((3, len(ps)), np.float64)
         ps_xy1[:2], ps_xy1[2] = points[ps].transpose(), 1
         ABC = np.concatenate((ab[ts], c[ts,:,None]), axis=2)
         B0 = ABC[:,0] @ ps_xy1 > - 1e-6
         B1 = ABC[:,1] @ ps_xy1 > - 1e-6
         B2 = ABC[:,2] @ ps_xy1 > - 1e-6
         w = np.where((B0 & B1) & B2)
         t, p = ts[w[0]], ps[w[1]]
         z = - (planes_d[t] + np.sum(planes_abc[t,:2] * points[p], axis=1)) / planes_abc[t,2]
         a = z.argsort()
         p_z[p[a]] = z[a]
      self.info.OK()
      self.info.OK()
      return p_z


   def is_in_surf(self, points):
      self.info.start("Searching if positions are occered in mesh area of {:n} points".format(len(points)))
      if self.params.tria_nodes is None: self.params.create_tria_nodes() 
      ab, c = self.params.get_tria_xy_sides_ab_c()
      p_surf = np.full(len(points), False, dtype=np.bool)
      buf = self.get_buf_points_triangles(points)
      self.info.start("Processing {:n} items of buffer".format(len(buf)))
      for ps, ts in buf:
         ps_xy1 = np.empty((3, len(ps)), np.float64)
         ps_xy1[:2], ps_xy1[2] = points[ps].transpose(), 1
         ABC = np.concatenate((ab[ts], c[ts,:,None]), axis=2)
         B0 = ABC[:,0] @ ps_xy1 > - 1e-6
         B1 = ABC[:,1] @ ps_xy1 > - 1e-6
         B2 = ABC[:,2] @ ps_xy1 > - 1e-6
         p_surf[ps[((B0 & B1) & B2).any(axis=0)]] = True         
      self.info.OK()
      self.info.OK()
      return p_surf

      
      

   # ******* 3D mass by elements **********************

   def get_e_type_buf_points_elems(self, t, points, corner=0):
      self.info.start("Building points & elements type {:n} buffer".format(t))
      r, m = self.params, self.mesh
      st_p, st_e = [[] for _ in range(8)], [[] for _ in range(8)]      
      s, nc = m.get_e_type_slice(t), m.E_type_nodes_count[t]
      buf = [(np.arange(0, len(points), dtype=np.int32), np.arange(s.start, s.stop, dtype=np.int32), r.x_0, r.y_0, r.z_0, r.x_range, r.y_range, r.z_range)]
      BUF = []      
      g = [None for _ in range(8)]
      w = (-1, -1, -1)
      while len(buf) > 0:         
         for s in st_p: s.clear()
         for s in st_e: s.clear()
         pts, elms, x0, y0, z0, xr, yr, zr = buf.pop(0)         
         if w != (xr, yr, zr):
            w = (xr, yr, zr)
            self.info("Ranges x: {:g}  y: {:g}  z: {:g}".format(*w))
         xh, yh, zh = xr/2, yr/2, zr/2
         xs, ys, zs = x0 + xh, y0 + yh, z0 + zh
         st_i = (points[pts,0] > xs) + 2*(points[pts,1] > ys) + 4*(points[pts,2] > zs)
         for p, i in zip(pts, st_i): st_p[i].append(p)
         n = m.nodes_xyz[m.e_nodes[elms,:nc]]
         D, U = n.min(axis=1) - corner, n.max(axis=1) + corner
         dx, dy, dz = D[:,0] < xs, D[:,1] < ys, D[:,2] < zs
         ux, uy, uz = U[:,0] >= xs, U[:,1] >= ys, U[:,2] >= zs
         for i in np.where((dx & dy) & dz)[0]: st_e[0].append(elms[i])
         for i in np.where((ux & dy) & dz)[0]: st_e[1].append(elms[i])
         for i in np.where((dx & uy) & dz)[0]: st_e[2].append(elms[i])
         for i in np.where((ux & uy) & dz)[0]: st_e[3].append(elms[i])            
         for i in np.where((dx & dy) & uz)[0]: st_e[4].append(elms[i])
         for i in np.where((ux & dy) & uz)[0]: st_e[5].append(elms[i])
         for i in np.where((dx & uy) & uz)[0]: st_e[6].append(elms[i])
         for i in np.where((ux & uy) & uz)[0]: st_e[7].append(elms[i])                           
         g[0] = xs - xh, ys - yh, zs - zh
         g[1] = xs     , ys - yh, zs - zh
         g[2] = xs - xh, ys     , zs - zh
         g[3] = xs     , ys     , zs - zh
         g[4] = xs - xh, ys - yh, zs
         g[5] = xs     , ys - yh, zs
         g[6] = xs - xh, ys     , zs
         g[7] = xs     , ys     , zs         
         for p, e, b in zip(st_p, st_e, g):
            c = len(p) * len(e)
            if c > self.OPTION_buf_points_elems_CRIT:               
               buf.append((np.array(p, dtype=np.int32), np.array(e, dtype=np.int32), *b, xh, yh, zh))
            elif c > 0: 
               BUF.append((np.array(p, dtype=np.int32), np.array(e, dtype=np.int32))) 
      self.info("Buffuer contains {:n} items.".format(len(BUF))) 
      self.info.OK()
      return BUF


   def get_points_e1(self, points):
      self.info.start("Searching elements of type 1 of {:n} points".format(len(points)))
      p_e = np.full(len(points), self.NO_E, dtype=np.int32)

      self.info.OK()
      return p_e


   def get_points_e2(self, points):
      self.info.start("Searching elements of type 2 of {:n} points".format(len(points)))
      abcd = self.params.get_e_type_planes_abcd(2)
      e2_half_cross = self.params.get_e_type_cross_sections(2)[:,None] / 2
      p_e = np.full(len(points), self.NO_E, dtype=np.int32)
      buf = self.get_e_type_buf_points_elems(2, points)           
      for ps, es in buf:
         ps_xyz1 = np.empty((4, len(ps)), np.float64)
         ps_xyz1[:3], ps_xyz1[3] = points[ps].transpose(), 1
         e2 = es - self.mesh.e_type_start_index[2]
         ABCD, w = abcd[e2], e2_half_cross[e2]
         B0 = np.absolute(ABCD[:,0] @ ps_xyz1) <= w
         B1 = ABCD[:,1] @ ps_xyz1 >= 0
         B2 = ABCD[:,2] @ ps_xyz1 >= 0
         B3 = ABCD[:,3] @ ps_xyz1 >= 0
         e, p = np.where(((B0 & B1) & B2) & B3)
         p_e[ps[p]] = es[e]         
         R = p_e[ps] == self.NO_E
         if any(R):
            B1 = ABCD[:,1] @ ps_xyz1[:,R] >= - w
            B2 = ABCD[:,2] @ ps_xyz1[:,R] >= - w
            B3 = ABCD[:,3] @ ps_xyz1[:,R] >= - w
            e, p = np.where(((B0[:,R] & B1) & B2) & B3)         
            p_e[ps[R][p]] = es[e] 
      self.info.OK()
      return p_e


   def get_points_e4(self, points):
      self.info.start("Searching elements of type 4 of {:n} points".format(len(points)))
      abcd = self.params.get_e_type_planes_abcd(4)    
      p_e = np.full(len(points), self.NO_E, dtype=np.int32) 
      buf = self.get_e_type_buf_points_elems(4, points)
      for ps, es in buf:
         ps_xyz1 = np.empty((4, len(ps)), np.float64)
         ps_xyz1[:3], ps_xyz1[3] = points[ps].transpose(), 1
         ABCD = abcd[es - self.mesh.e_type_start_index[4]]         
         B0 = ABCD[:,0] @ ps_xyz1 > - 1e-6
         B1 = ABCD[:,1] @ ps_xyz1 > - 1e-6
         B2 = ABCD[:,2] @ ps_xyz1 > - 1e-6
         B3 = ABCD[:,3] @ ps_xyz1 > - 1e-6
         e, p = np.where(((B0 & B1) & B2) & B3)
         p_e[ps[p]] = es[e]
      self.info.OK()
      return p_e

 
   def get_points_elements(self, points, E1=True, E2=True, E4=True):
      self.info.start("Getting elements of points")
      p_e = np.full(len(points), self.NO_E, np.int32)
      if E1:
         p_e[:] = self.get_points_e1(points)
      if E2:
         A = p_e == self.NO_E
         p_e[A] = self.get_points_e2(points[A])
      if E4:
         A = p_e == self.NO_E
         p_e[A] = self.get_points_e4(points[A])
      self.info.OK()
      return p_e



