import numpy as np

from .mesh_toolkit import MeshToolKit
from .mesh import Mesh

class DataToolKit(MeshToolKit):

   def __init__(self, params, DEM, bands):
      super().__init__(params=params, DEM=DEM, bands=bands)
      self.e_data = {}
      self.raw_ep_ef_sp_sf = {}


   def update_method(self):
      self.e_data.clear()
      self.raw_ep_ef_sp_sf.clear()


   def get_selected_e_data_ref(self, file_names=None, names=None, times=None, times_indices=None):
      r = []
      for k in self.e_data:
         f = file_names is None
         if not f: f = k[0] in file_names             
         n = names is None
         if not n: n = k[1] in names
         t = times is None
         if not t: t = k[2] in times             
         i = times_indices is None
         if not i: i = k[3] in times_indices
         if f and n and t and i: r.append(k)
      return r


   # ************ Data loading *************************

   def load_element_data_from_same_mesh(self, file_name, **selection):
      self.info.start("Loading element data from identical mesh file")
      e_data = self.mesh.try_load_msh_element_data(file_name=file_name, **selection)
      for k in e_data: self.e_data[k] = e_data[k]
      refs = sorted(e_data, key=lambda r: (r[1], r[3]))
      self.info.OK()
      return refs


   def load_element_data_from_any_mesh(self, file_name, **selection):
      self.info.start("Loading data from output msh file")
      m_out = Mesh(self.info)
      m_out.load_msh(file_name, None)      
      T_out = self.params.calc_e_T(mesh=m_out)
      e = self.bands.get_points_elements(T_out)
      e_data_out = m_out.try_load_msh_element_data(file_name=file_name, **selection)
      m = self.mesh
      for k, d_out in e_data_out.items():
         if len(d_out.shape) == 1: d = np.full(m.get_e_count(), np.nan, np.float64)
         else: d = np.full((m.get_e_count(), d_out.shape[1]), np.nan, np.float64)
         d[e] = d_out
         self.e_data[k] = d
      refs = sorted(e_data_out, key=lambda r: (r[1], r[3]))
      self. info.OK()
      return refs



   def load_water_flow_raw(self, file_name):
      self.info.start("Loading water flow raw file '" + file_name + "'")
      m = self.mesh
      self.raw_ep_ef_sp_sf.clear()
      id_e = {id:e for e, id in enumerate(m.e_gmsh_ids)}
      with open(file_name) as f:
         f.readline()
         f.readline()         
         s = f.readline().strip()
         while s == "$FlowField":
            T = float(f.readline().split("=")[1])
            d = np.full((m.get_e_count(),12), np.nan, np.float64)            
            for _ in range(int(f.readline())):
               s = f.readline().split()
               i = id_e[int(s.pop(0))]
               if (c:=s.pop(4)) == "4": d[i] = s
               elif c == "3": d[i,:10] = s
               else: c[i,:8] = s                       
            self.raw_ep_ef_sp_sf[T] = (d[:,0], d[:,1:4], d[:,4:8], d[:,8:])
            f.readline()
            self.info("Time {:g} loaded".format(T))
            s = f.readline().strip()
         if len(self.raw_ep_ef_sp_sf) == 0:
            raise self.ToolKitException("No $FlowField section was found.")
      self.info.OK()
      return self.raw_ep_ef_sp_sf
      

   # ************ data info **************************
   
   class DataInfo: pass

   def get_e_data_info(self, data_refs=None):
      self.info.start("Getting elements data info")
      r_d = {}
      for r, d in self.e_data.items():
         if data_refs is not None:
            if r not in data_refs: continue
         n = self.DataInfo()
         n.no_value_count = np.sum(np.isnan(d))
         n.dim = 1 if len(d.shape) == 1 else d.shape[1]
         n.min = np.nanmin(d)         
         n.max = np.nanmax(d)
         n.mean = np.nanmean(d)
         n.median = np.nanmedian(d)         
         n.protocol = [
            ("Value dimension", n.dim),
            ("Minimum", "{:g}".format(n.min)),
            ("Maximum", "{:g}".format(n.max)),
            ("Mean", "{:g}".format(n.mean)),
            ("Median", "{:g}".format(n.median))            
         ]
         if n.no_value_count > 0:
            n.protocol.append(("No value count", n.no_value_count))
         r_d[r] = n
      self.info.OK()
      return r_d


   # ************* Data manipulation *****************

   def set_e_data_boundary_e2_as_e4_neighbours(self, data_refs, e2_selection):
      self.info.start("Setting data of boundary elements type 2 as neighbours elements type 4")
      m = self.mesh
      if e2_selection is None:
         sc = m.get_e_type_slice(2)
         elems = np.arange(sc.start, sc.stop, dtype=np.int32)
      else: elems = e2_selection
      e_nbs = self.params.get_e2_neighbours_e4_by_plane(e2_selection=e2_selection)
      a = np.sum(e_nbs != - 1, axis=1)
      if n:=np.sum(a == 0): raise self.ToolKitException("{:n} elements has no neighbour.".format(n))
      if n:=np.sum(a == 2): raise self.ToolKitException("{:n} elements are not boundary (They has two neighbours.".format(n))
      status = []
      for r in (self.e_data if data_refs is None else data_refs):
         self.e_data[r][elems] = self.e_data[r][e_nbs[:,0]]       
         status.append((r, len(elems) - np.sum(np.isnan(self.e_data[r][elems])), len(elems)))         
         self.info("Name: '{:s}',  Time: {:s},  Step: {:n},  setted elements: {:n} of {:n}".format(*r[1:], *status[-1][1:]))
      self.info.OK()
      return status


   def create_scalar_e_data_from_3D_vectors(self, vector_e_data_refs, x_part=True, y_part=True, z_part=True, vert_angle=True, horiz_angle=True, size=True):
      self.info.start("Creating scalar elements data from 3D vectors e data")
      for r in vector_e_data_refs:
         d = self.e_data[r]
         if x_part:
            R = (r[0], r[1]+"_X", r[2], r[3])
            self.e_data[R] = np.array(d[:,0], dtype=np.float64)
         if y_part:
            R = (r[0], r[1]+"_Y", r[2], r[3])
            self.e_data[R] = np.array(d[:,1], dtype=np.float64)
         if z_part:
            R = (r[0], r[1]+"_Z", r[2], r[3])
            self.e_data[R] = np.array(d[:,2], dtype=np.float64)
         if size:
            R = (r[0], r[1]+"_SIZE", r[2], r[3])
            self.e_data[R] = np.sum(d**2, axis=1)**0.5
         if vert_angle:
            R = (r[0], r[1]+"_VERTICAL_ANGLE", r[2], r[3])
            self.e_data[R] = np.asin(d[:,2] / np.sum(d**2, axis=1)**0.5) / np.pi * 180
         if horiz_angle:
            R = (r[0], r[1]+"_HORIZONTAL_ANGLE", r[2], r[3])
            n = np.sum(d[:,:2]**2, axis=1)**0.5
            a = np.acos(d[:,0] / n) / np.pi * 180            
            B = d[:,1] / n < 0
            a[B] = 360 - a[B]
            self.e_data[R] = a
      self.info.OK()
      

   def calc_new_e_data(self, data_refs, new_name, formula):
      self.info.start("Calculating new elements data")
      if len(data_refs) > 0:
         if len({r[1] for r in data_refs}) > 1:
            raise self.ToolKitException("Quantities for computing new quantity '"+ new_name +"' have not identical names.")
         is_refs = True
      else:          
         if "data" in formula:
            raise self.ToolKitException("Formula can not contain variable 'data', because no element data were selected.")
         data_refs = [(None, None, '0', 0)]
         is_refs = False
      is_T = "T" in formula
      is_TdepthDEM = "TdepthDEM" in formula
      is_TdepthTOP = "TdepthTOP" in formula
      if is_T: 
         e_T = self.params.get_e_T()         
      if is_TdepthDEM:
         e_depthDEM = self.DEM.get_points_elevations(e_T[:,:2]) - e_T[:,2]
      if is_TdepthTOP:                                                     
         e_depthTOP = self.bands.get_surf_z(e_T[:,:2]) - e_T[:,2]
      new_refs = []
      for ref in data_refs:
         if is_refs:
            data = self.e_data[ref]
            is_vector = len(data.shape) == 2
         else:
            is_vector = False
         if is_T:
            if not is_vector:
               Tx, Ty, Tz = e_T[:,0], e_T[:,1], e_T[:,2]
               if is_TdepthDEM: TdepthDEM = e_depthDEM
               if is_TdepthTOP: TdepthTOP = e_depthTOP
            else:
               Tx, Ty, Tz = e_T[:,0,None], e_T[:,1,None], e_T[:,2,None]
               if is_TdepthDEM: TdepthDEM = e_depthDEM[:,None]
               if is_TdepthTOP: TdepthTOP = e_depthTOP[:,None]               
         new_data = eval(formula)            
         new_ref = ("Calculated", new_name, *ref[2:])
         self.e_data[new_ref] = new_data
         new_refs.append(new_ref)
         self.info("New quantity: {:s}, time: {:s}, time step: {:n}".format(*new_ref[1:]))
      self.info.OK()
      return new_refs


