import numpy as np

from .mesh_toolkit import MeshToolKit
from .mesh import NO_NODE_INDEX, Physical

class MeshEditToolKit(MeshToolKit):

   def __init__(self, DEM, params, bands):
      super().__init__(DEM=DEM, params=params, bands=bands)


   def change_top_nodes_z_by_DEM(self, desc_file_name):
      self.info.start("Changing top nodes z by DEM")      
      m = self.mesh
      nodes = m.get_nodes_of_physicals(tp:=self.params.OPTIONS_top_boundary_physicals_ids)
      z = self.DEM.get_points_elevations(m.nodes_xyz[nodes,:2])
      delta = z - m.nodes_xyz[nodes,2]
      m.nodes_xyz[nodes,2] += delta
      if desc_file_name is not None:
         m.save_msh_with_selected_physicals(desc_file_name, tp)
         m.append_nodes_data(desc_file_name, "delta_z", "0", 0, m.nodes_id[nodes], delta)
      self.info.OK()


   def change_nodes_z_from_level_by_DEM(self, z_level, desc_file_name):
      self.info.start("Changing nodes z by DEM of z level {:g} m".format(z_level))
      m = self.mesh
      nodes = m.nodes_xyz
      B = nodes[:,2] >= z_level      
      self.info("Seleted depth contains {:n} nodes.".format(np.sum(B)))
      elv_DEM = self.DEM.get_points_elevations(nodes[B,:2])
      elv_m = self.bands.get_surf_z(nodes[B,:2])
      delta = (elv_DEM - elv_m) * (1 - (elv_m - nodes[B,2]) / (elv_m - z_level))
      nodes[B,2] += delta
      if desc_file_name is not None:
         m.save_msh(desc_file_name)
         m.append_nodes_data(desc_file_name, "delta_z", "0", 0, m.nodes_id[B], delta)
      self.info.OK()


   def remove_physicals(self, physicals_ids):
      self.info.start("Removing physicals with ids: "+(len(physicals_ids)*"{:n}").format(*physicals_ids))
      m = self.mesh      
      del_e = np.concatenate([m.physicals[id].elements for id in physicals_ids])
      # *********** delete elements **************      
      e_phs = np.delete(m.get_e_physicals_ids(), del_e)
      m.e_nodes = np.delete(m.e_nodes, del_e, axis=0)
      m.e_order_ids = np.delete(m.e_order_ids, del_e)
      m.e_gmsh_ids = np.delete(m.e_gmsh_ids, del_e)            
      # *********** physicals ******************
      for id in physicals_ids: m.physicals.pop(id)
      for p in m.physicals.values():
         p.elements = np.where(e_phs == p.id)
      # **** new e type start index elements in physicals *****
      m.e_type_start_index = [i - np.sum(del_e < i) for i in m.e_type_start_index]
      # ******** delete single nodes ***************        
      used_nodes = np.unique(m.e_nodes[m.e_nodes!=NO_NODE_INDEX])
      used_nodes.sort()
      d = {old_i:new_i for new_i, old_i in enumerate(np.arange(len(m.nodes_id))[used_nodes])}
      f = np.vectorize(pyfunc=lambda old_i: d[old_i], otypes=[np.int32])
      for t, c in enumerate(m.E_type_nodes_count):
         if m.get_e_type_count(t) > 0: 
            m.e_nodes[s,:c] = f(m.e_nodes[s:=m.get_e_type_slice(t),:c])
      m.nodes_id = m.nodes_id[used_nodes]
      m.nodes_xyz = m.nodes_xyz[used_nodes]
      self.info.OK()
     

   def merge_physicals(self, new_name, new_id, physicals_ids):
      self.info.start("Merging physicals to new physical with id {:n} and name '{:s}'".format(new_id, new_name))
      p = Physical(new_id)
      p.name = new_name
      p.dim = self.mesh.physicals[physicals_ids[0]].dim
      elems = []
      for id in physicals_ids:
         elems.append(self.mesh.physicals.pop(id).elements)
      p.elements = np.concatenate(elems)
      self.mesh.physicals[new_id] = p
      self.info.OK()


   def change_physical_id_name(self, id, new_id, new_name):
      self.info.start("Changing id and name of physical id {:n}".format(id))
      s = self.mesh.physicals
      if new_id != id:
         if new_id in s:
            raise self.ToolKitException("ID {:n} still exists for name '{:s}'.".format(new_id, s[new_id].name))
         p = s.pop(id)
         id = p.id = new_id
         s[id] = p
         self.info("New ID: {:n}".format(id))
      if new_name != s[id].name:
         if new_name in {p.name for p in s.values()}:
            raise self.ToolKitException("Name '{:s}' still exists.".format(new_name))
         s[id].name = new_name
         self.info("New name: '{:s}'".format(new_name))
      self.info.OK()





