import os, itertools
import numpy as np
# import h5py as h5


NO_NODE_INDEX = 0x7FFFFFFF


class Physical:

   def __init__(self, id):
      self.id = id
      self.dim = None
      self.name = None
      self.elements = []



class Mesh_Err:
   class File_Open_Error(Exception): pass
   class No_Section_Error(Exception): pass
   class Physicals_Load_Error(Exception): pass
   class ElementData_Load_Error(Exception): pass


class Mesh:

   NO_NODE_INDEX = NO_NODE_INDEX
   E_type_nodes_count = [0, 2, 3, 4, 4, 8, 6, 5]
   E_type_names = ["virtual", "line", "triangle", "square", "tetrahedron", "hexahedron", "wedge", "pyramid"]
   pflotran_ascii_ugi_file_ext = "ugi"
   pflotran_ascii_regions_file_ext = "r"
   pflotran_ascii_boundary_region_file_ext = "ss"

   def __init__(self, info):
      self.info = info

   
   def get_e_type(self, e):
      for t in range(7, -1, -1):
         if e >= self.e_type_start_index[t]: return t

  
   # ************** LOAD MSH ***********************
        
   def _open_file(self, file_name):
      try:
         f = open(file_name, "r")
      except FileNotFoundError:
         raise Mesh_Err.File_Open_Error("File '" + file_name + "' opening error")
      return f


   def _goto_section(self, f, section_name, raise_exception=True):
      s0 = section_name[0]
      for s in f:
         if s0 in s:
            if s[1:-1] == section_name[1:]: return True
      if raise_exception: 
         raise Mesh_Err.No_Section_Error("Mesh file: '{:s}'\nSection '{:s}' was not found.".format(f.name, section_name))
      else:
         return False


   def _load_nodes(self):
      self.info.start("Loading nodes from file '{:s}'".format(self.msh_file))
      with self._open_file(self.msh_file) as f:
         self._goto_section(f, "$Nodes")
         n = int(f.readline())
         self.nodes_id = id = np.empty(n, np.int32)
         self.nodes_xyz = xyz = np.empty((n, 3), np.float64)
         for i in range(n):
            p = f.readline().split()
            id[i], xyz[i] = p[0], p[1:]
      self.info.OK("{:d} nodes were loaded.".format(n))


   def _load_elements_with_physicals(self):
      self.info.start("Loading elements with physicals from file '{:s}'".format(self.msh_file))
      # ***** load lines with elemennts from mesh file ************
      gen_e_order_id = itertools.count(0)     
      with self._open_file(self.msh_file) as f:
         self._goto_section(f, "$Elements")
         n = int(f.readline())
         e_lines = [(next(gen_e_order_id), *map(int, f.readline().split())) for _ in range(n)]
      # ******** sort lines by elements types *************************
      e_lines.sort(key=lambda line: line[2])
      # ****** elements types start index ********************
      e_types = list(map(lambda p: p[2], e_lines))
      c = [e_types.count(t) for t in range(8)]
      self.e_type_start_index = [sum(c[:t]) for t in range(8)]
      # ******** create elements structure ******************
      self.e_order_ids = np.fromiter(map(lambda p: p[0], e_lines), np.int32)
      self.e_gmsh_ids = np.fromiter(map(lambda p: p[1], e_lines), np.int32)
      # ******** create nodes elements **********************
      self.e_nodes = np.full((len(e_lines), max(n for i, n in enumerate(self.E_type_nodes_count) if c[i] > 0)), NO_NODE_INDEX, np.int32)
      nodes_dict = {id:i for i, id in enumerate(self.nodes_id)}
      for t, nc in enumerate(self.E_type_nodes_count):
         if c[t] > 0:
            sc = self.get_e_type_slice(t)
            w = self.e_nodes[sc,:nc]
            for i, p in enumerate(e_lines[sc]):
               w[i] = [nodes_dict[id] for id in p[(4+p[3]):]]
      # ***** physicals **********************************
      self.physicals = {}
      for e, p in enumerate(e_lines):
         (self.physicals[p[4]] if p[4] in self.physicals else self.physicals.setdefault(p[4], Physical(p[4]))).elements.append(e)
      for p in self.physicals.values():
         p.elements = np.fromiter(p.elements, np.int32)
      # ***********************************************
      for t, n in enumerate(c):
         if n > 0: self.info("Number of type {:n}: {:n}".format(t, n))
      self.info.OK("{:n} elements in {:n} physicals were loaded.".format(len(self.e_nodes), len(self.physicals)))


   def _set_physicals_name(self):
      self.info.start("Setting physical names")
      if self.msh_file_physicals_name is None:
         self.info("Setting dimensions only without $PhysicalNames")
         for phs in self.physicals.values():
            e = next(e for e in phs.elements)
            if e < self.e_type_start_index[2]: phs.dim = 1
            elif e < self.e_type_start_index[4]: phs.dim = 2
            else: phs.dim = 3
            phs.name = "{:d}D_{:d}".format(phs.dim, phs.id)
      else:
         with self._open_file(self.msh_file_physicals_name) as f:
            self.info("Loading physicals names from file '{:s}'".format(f.name))
            self._goto_section(f, "$PhysicalNames", raise_exception=True)
            for _ in range(int(f.readline())):
               p = f.readline().replace('"','').split()
               phs = self.physicals.get(int(p[1]), None)
               if phs is not None:
                  phs.dim, phs.name = int(p[0]), " ".join(p[2:])
            for phs in self.physicals.values():
               if phs.name is None:
                  raise Mesh_Err.Physicals_Load_Error("Mesh file: '{:s}'\nPhysical with id {:d} has no item in $PhysicalNames section.".format(phs.id, self.f.name))
      self.info.OK()


   def _test_physical_elems_type(self):
      self.info.start("Testing element types in physicals")
      dim_etyp = {1: {1}, 2: {2,3}, 3: {4,5,6,7}}
      for phs in self.physicals.values():
         u = [(self.e_id, typ) for e in phs.elements if (typ:=self.get_e_type(e)) not in dim_etyp[phs.dim]]
         if len(u) > 0:
            raise Mesh_Err.Physicals_Load_Error("Mesh file: '{:s}'\nElement {:d} of type {:d} canot be contain in physical {:d} of dimension {:d}.".format(self.f.name, u[0][0], u[0][1], phs.id, phs.dim))
      self.info.OK()


   def load_msh(self, msh_file, msh_file_physicals_name):
      self.info.start("Loading mesh of gmsh format")
      self.msh_file, self.msh_file_physicals_name = msh_file, msh_file_physicals_name
      self._load_nodes()
      self._load_elements_with_physicals()
      self._set_physicals_name()
      self._test_physical_elems_type()
      self.info.OK()


   def load_msh_element_data_references(self, file_name=None):
      if file_name is None: file_name = self.msh_file
      self.info.start("Loading $ElementData referenceds from file '{:s}'".format(file_name))                  
      with self._open_file(file_name) as f:
         refs, ed_raise = [], True
         while self._goto_section(f, "$ElementData", raise_exception=ed_raise):
            ed_raise = False
            # ******** strings ********************
            n = int(f.readline())         
            f_names = [f.readline().strip()[1:-1] for _ in range(n)]
            # ******** floats *********************
            n = int(f.readline())
            if n != 1: raise Mesh_Err.ElementData_Load_Error("Mesh file: '{:s}'\nHead of $ElementData must have just one float (time) value.".format(self.f.name))           
            f_time = f.readline().strip()
            # ******** integers *******************
            n = int(f.readline())
            if n != 3: raise self.ElementData_Load_Error("Mesh file: '{:s}'\nHead of $ElementData must have just 3 integer values.".format(self.f.name))
            f_time_index = int(f.readline())
            f_parts_count = int(f.readline())
            if f_parts_count > 1 and len(f_names) > 1: raise self.ElementData_Load_Error("Mesh file: '{:s}'\nHead of $ElementData must have just 1 string value or just 1 data part.".format(self.f.name))
            for s in f_names:
               refs.append((file_name, s, f_time, f_time_index))         
      self.info.OK()
      return refs
      


   def try_load_msh_element_data(self, file_name=None, names=None, times=None, times_indices=None):
      if file_name is None: file_name = self.msh_file
      self.info.start("Loading $ElementData from file '{:s}'".format(file_name))            
      id_e = {id:e for e, id in enumerate(self.e_gmsh_ids)}
      with self._open_file(file_name) as f:
         e_data, ed_raise = {}, True
         while self._goto_section(f, "$ElementData", raise_exception=ed_raise):
            ed_raise = False
            # ******** strings ********************
            n = int(f.readline())
            if names is None:
               f_names = [f.readline().strip()[1:-1] for _ in range(n)]
            else: f_names = [s for _ in range(n) if (s:=f.readline().strip()[1:-1]) in names]
            if (f_names_count := len(f_names)) == 0: continue         
            # ******** floats *********************
            n = int(f.readline())
            if n != 1: raise Mesh_Err.ElementData_Load_Error("Mesh file: '{:s}'\nHead of $ElementData must have just one float (time) value.".format(self.f.name))           
            f_time = f.readline().strip()
            if times is not None:
               if f_time not in times: continue
            # ******** integers *******************
            n = int(f.readline())
            if n != 3: raise self.ElementData_Load_Error("Mesh file: '{:s}'\nHead of $ElementData must have just 3 integer values.".format(self.f.name))
            f_time_index = int(f.readline())
            if times_indices is not None:
               if f_time_index not in times_indices: continue
            f_parts_count = int(f.readline())
            if f_parts_count > 1 and f_names_count > 1: raise self.ElementData_Load_Error("Mesh file: '{:s}'\nHead of $ElementData must have just 1 string value or just 1 data part.".format(self.f.name))            
            # *********** data ********************
            e_not_found = 0
            f_values_count = int(f.readline())
            if f_names_count == 1 and f_parts_count == 1:
               lines = [int(p[0]) if (p:=f.readline().strip().partition(" "))[2] == "0" else (int(p[0]), p[2]) for _ in range(f_values_count)]
               R = (file_name, f_names[0], f_time, f_time_index)
               d = e_data[R] = np.full(self.get_e_count(), np.nan, np.float64)
               for p in lines: 
                  if type(p) is int:
                     if (a:=id_e.get(p)) is not None: d[a] = 0
                     else: e_not_found += 1
                  elif (a:=id_e.get(p[0])) is not None: d[a] = p[1]
                  else: e_not_found += 1
               self.info("{:d}. name: {:s}, time: {:s}, time index: {:n} was loaded.".format(len(e_data), *R[1:]))
            elif f_names_count == 1 and f_parts_count > 1:
               s0 = " ".join(f_parts_count * ["0"])
               lines = [int(p[0]) if (p:=f.readline().strip().partition(" "))[2] == s0 else (int(p[0]), *p[2].split()) for _ in range(f_values_count)]
               R = (file_name, f_names[0], f_time, f_time_index)
               d = e_data[R] = np.full((self.get_e_count(), f_parts_count), np.nan, np.float64)
               for p in lines: 
                  if type(p) is int:
                     if (a:=id_e.get(p)) is not None: d[a] = 0
                     else: e_not_found += 1
                  elif (a:=id_e.get(p[0])) is not None: d[a] = p[1:]
                  else: e_not_found += 1
               self.info("{:d}. name: {:s}, time: {:s}, time index: {:n} was loaded.".format(len(e_data), *R[1:]))
            elif f_names_count > 1:
               s0 = " ".join(f_names_count * ["0"])
               lines = [int(p[0]) if (p:=f.readline().strip().partition(" "))[2] == s0 else (int(p[0]), *p[2].split()) for _ in range(f_values_count)]
               for i, f_name in enumerate(f_names, 1):
                  R = (file_name, f_name, f_time, f_time_index)
                  d = e_data[R] = np.full(self.get_e_count(), np.nan, np.float64)
                  for p in lines:
                     if type(p) is int:
                        if (a:=id_e.get(p)) is not None: d[a] = 0
                        else: e_not_found += 1
                     elif (a:=id_e.get(p[0])) is not None: d[a] = p[i]
                     else: e_not_found += 1                           
                  self.info("{:d}. name: {:s}, time: {:s} time index {:n} was loaded.".format(len(e_data), *R[1:]))        
            if e_not_found > 0:
               self.info("WARNING !!! {:n} element IDs with data were not found in mesh.".format(e_not_found))
            if names is not None and (times is not None or times_indices is not None):
               if len(names) * ((0 if times is None else len(times)) + (0 if times_indices is None else len(times_indices)) ) == len(e_data): break            
      if len(e_data) == 0:
         raise Mesh_Err.ElementData_Load_Error("No element data were fount with given quantities names and times.")
      self.info.OK()
      return e_data
  


   # *********** SAVE MSH ***************************

   def save_msh(self, file_name):
      self.info.start("Saving mesh to file: '" + file_name+ "'")
      with open(file_name, "w") as f:
         f.write("$MeshFormat\n2.2 0 8\n$EndMeshFormat\n")
         phs = sorted(self.physicals.values(), key=lambda p: p.id)
         f.write("$PhysicalNames\n{:n}\n".format(len(phs)))         
         f.writelines(map(lambda p: "{} {} \"{:s}\"\n".format(p.dim, p.id, p.name), phs))
         f.write("$EndPhysicalNames\n")
         self.info("Number of physicals: {:n}".format(len(phs)))
         f.write("$Nodes\n{:n}\n".format(len(self.nodes_id)))
         s = "{} {} {} {}\n"
         f.writelines(s.format(id, *xyz.tolist()) for id, xyz in zip(self.nodes_id, self.nodes_xyz))
         f.write("$EndNodes\n")
         self.info("Number of Nodes: {:n}".format(len(self.nodes_id)))
         f.write("$Elements\n{:n}\n".format(self.get_e_count()))
         e_phs = self.get_e_physicals_ids()
         s = ["{} "+str(t)+" 2 {} 0" + c*" {}" + "\n" for t, c in enumerate(self.E_type_nodes_count)]
         e_lines = []
         for t, nc in enumerate(self.E_type_nodes_count):
            if self.get_e_type_count(t) > 0:
               sc, st = self.get_e_type_slice(t), s[t]
               e_lines.extend((p0, st.format(p1, p2, *p3.tolist())) for p0, p1, p2, p3 in zip(self.e_order_ids[sc].tolist(), self.e_gmsh_ids[sc].tolist(), e_phs[sc].tolist(), self.nodes_id[self.e_nodes[sc,:nc]]))
         e_lines.sort(key = lambda p: p[0])
         f.writelines(p for _, p in e_lines)
         f.write("$EndElements\n")
         self.info("Number of elements: {:n}".format(self.get_e_count()))
      self.info.OK()


   def save_msh_with_selected_physicals(self, file_name, physicals_ids):
      self.info.start("Saving mesh with selected physicals to file: '" + file_name+ "'")      
      with open(file_name, "w") as f:
         f.write("$MeshFormat\n2.2 0 8\n$EndMeshFormat\n")       
         physicals = sorted((self.physicals[id] for id in physicals_ids), key=lambda p: p.id)
         f.write("$PhysicalNames\n{:n}\n".format(len(physicals)))         
         f.writelines(map(lambda p: "{} {} \"{:s}\"\n".format(p.dim, p.id, p.name), physicals))
         f.write("$EndPhysicalNames\n")
         self.info("Number of physicals: {:n}".format(len(physicals)))
         nodes = self.get_nodes_of_physicals(physicals_ids)
         f.write("$Nodes\n{:n}\n".format(len(nodes)))
         s = "{} {} {} {}\n"
         f.writelines(s.format(id, *xyz.tolist()) for id, xyz in zip(self.nodes_id[nodes], self.nodes_xyz[nodes]))
         f.write("$EndNodes\n")
         self.info("Number of Nodes: {:n}".format(len(nodes)))
         elems = self.get_elements_of_physicals(physicals_ids)
         f.write("$Elements\n{:n}\n".format(len(elems)))
         e_phs = self.get_e_physicals_ids()
         s = ["{} "+str(t)+" 2 {} 0" + c*" {}" + "\n" for t, c in enumerate(self.E_type_nodes_count)]
         e_lines = []
         for t, nc in enumerate(self.E_type_nodes_count):
            if self.get_e_type_count(t) > 0:
               et, st = self.choose_e_type(t, elems), s[t]
               e_lines.extend((p0, st.format(p1, p2, *p3.tolist())) for p0, p1, p2, p3 in zip(self.e_order_ids[et].tolist(), self.e_gmsh_ids[et].tolist(), e_phs[et].tolist(), self.nodes_id[self.e_nodes[et,:nc]]))
         e_lines.sort(key = lambda p: p[0])
         f.writelines(p for _, p in e_lines)
         f.write("$EndElements\n")
         self.info("Number of elements: {:n}".format(len(elems)))
      self.info.OK()


   def save_msh_with_selected_elements(self, file_name, elems):
      self.info.start("Saving mesh with selected elements to file: '" + file_name+ "'")      
      with open(file_name, "w") as f:
         f.write("$MeshFormat\n2.2 0 8\n$EndMeshFormat\n")       
         physicals = sorted((p for p in self.physicals.values() if np.isin(elems, p.elements, assume_unique=True).any()), key=lambda p: p.id)        
         f.write("$PhysicalNames\n{:n}\n".format(len(physicals)))         
         f.writelines(map(lambda p: "{} {} \"{:s}\"\n".format(p.dim, p.id, p.name), physicals))
         f.write("$EndPhysicalNames\n")
         self.info("Number of physicals: {:n}".format(len(physicals)))
         nodes = np.unique(self.e_nodes[elems].reshape(-1))
         nodes = nodes[nodes != NO_NODE_INDEX]
         f.write("$Nodes\n{:n}\n".format(len(nodes)))
         s = "{} {} {} {}\n"
         f.writelines(s.format(id, *xyz.tolist()) for id, xyz in zip(self.nodes_id[nodes], self.nodes_xyz[nodes]))
         f.write("$EndNodes\n")
         self.info("Number of Nodes: {:n}".format(len(nodes)))        
         f.write("$Elements\n{:n}\n".format(len(elems)))
         e_phs = self.get_e_physicals_ids()
         s = ["{} "+str(t)+" 2 {} 0" + c*" {}" + "\n" for t, c in enumerate(self.E_type_nodes_count)]
         e_lines = []
         for t, nc in enumerate(self.E_type_nodes_count):
            if self.get_e_type_count(t) > 0:
               et, st = self.choose_e_type(t, elems), s[t]
               e_lines.extend((p0, st.format(p1, p2, *p3.tolist())) for p0, p1, p2, p3 in zip(self.e_order_ids[et].tolist(), self.e_gmsh_ids[et].tolist(), e_phs[et].tolist(), self.nodes_id[self.e_nodes[et,:nc]]))
         e_lines.sort(key = lambda p: p[0])
         f.writelines(p for _, p in e_lines)
         f.write("$EndElements\n")
         self.info("Number of elements: {:n}".format(len(elems)))
      self.info.OK()


   def append_nodes_data(self, file_name, name, time, time_index, ids, data, num_format="G"):
      self.info.start("Adding nodes data to mesh file: '" + file_name+ "'")
      with open(file_name, "a") as f:
         c = 0 if len(data.shape) == 1 else data.shape[1]
         s = "{}" + (1 if c==0 else c) * (" {}" if num_format == "" else (" {:" + num_format + "}")) + "\n"
         f.write('$NodeData\n1\n"{:s}"\n1\n{:s}\n3\n{:n}\n{:n}\n{:n}\n'.format(name, time, time_index, 1 if c==0 else c, data.shape[0]))
         if c == 0: f.writelines(s.format(*id_value) for id_value in zip(ids, data))
         else: f.writelines(s.format(id, *values) for id, values in zip(ids, data))
         f.write("$EndNodeData\n")  
      self.info.OK()


   def append_element_data_as_one_name(self, file_name, name, time, time_index, elems, data, num_format="G"):
      self.info.start("Adding element data (quantity: {:s},  time: {:s},  step: {:n}) to mesh file: '{:s}'".format(name, time, time_index, file_name))
      c = 0 if len(data.shape) == 1 else data.shape[1]
      s = "{}" + (1 if c==0 else c) * (" {}" if num_format == "" else (" {:" + num_format + "}")) + "\n"            
      B = np.logical_not(np.isnan(data))
      if c != 0: B = np.all(B, axis=1)
      elems, data = elems[B], data[B]
      a = np.argsort(self.e_order_ids[elems])
      ids = self.e_gmsh_ids[elems[a]]
      d = data[a]      
      with open(file_name, "a") as f:
         f.write('$ElementData\n1\n"{:s}"\n1\n{:s}\n3\n{:n}\n{:n}\n{:n}\n'.format(name, time, time_index, 1 if c==0 else c, data.shape[0]))
         if c == 0: f.writelines(s.format(*id_value) for id_value in zip(ids, d))
         else: f.writelines(s.format(id, *values) for id, values in zip(ids, d))
         f.write("$EndElementData\n")  
      self.info.OK()


   # ********** PfLOTRAN ASCII files management **************

   def create_pflotran_regions_files_list(self, file_dir, file_prefix, file_ext):
      f_dir = file_dir if file_dir[-1] in ("/","\\") else file_dir + "/"
      m = os.listdir(file_dir)     
      f_n = [(f_dir + f, f) for p in m if (f:=p.split("/")[-1]).startswith(file_prefix) and f.endswith("."+file_ext)]
      files = [s for s, _ in f_n]
      names = [s for _, s in f_n]
      return files, names


   def load_pflotran_ascii_ugi_and_boundary_regions(self, ugi_file, br_files, br_names):
      self.info.start("Loading PFloTran unstrucured grid from ASCII files")
      self.pflotran_ascii_ugi_file = ugi_file
      # **** no element type *************
      e_lines = [(0, 0, 0)]
      e_id = itertools.count(1)
      # ***** loading mass elements and nodes *************************
      t_dict = {"T":4, "H":5, "W":6, "P":7}
      with open(ugi_file) as f:
         e_count, n_count = tuple(map(int, f.readline().split()))
         e_lines.extend((next(e_id), t_dict[(p:=f.readline().split())[0]], *tuple(map(int, p[1:]))) for _ in range(e_count))
         self.nodes = [Node(id, tuple(map(float, f.readline().split()))) for id in range(1, n_count+1)]
         self.info("Grid from file '" + ugi_file + "' was loaded.")
      # **** loading boundary elements and boundary regions **
      self.physicals = {}
      t_dict = {"T":2, "Q":3}
      for phs_id, file_name in enumerate(br_files):
         phs = Physical(phs_id)
         phs.dim = 2
         phs.name = file_name if br_names is None else br_names[phs_id]
         with open(file_name) as f:
            f.readline()
            phs.elements = [(next(e_id), t_dict[(p:=line.split())[0]], *tuple(map(int, p[1:]))) for line in f]
         e_lines.extend(phs.elements)
         self.physicals[phs.name] = phs
         self.info("Boundary region from file '" + file_name + "' was loaded.")
      # ***** sorting by element types *********************
      e_lines.sort(key=lambda s: s[1])
      # ****** set physicals elements by e_lines *******************
      e_line2e = {e_line:e for e, e_line in enumerate(e_lines)}
      for p in self.physicals.values():
         p.elements = np.fromiter(map(lambda e_line: e_line2e[e_line], p.elements), np.int32)
      # ****** slice by elements type ************************
      e_types = list(map(lambda p: p[1], e_lines))
      c = self.e_type_count = [e_types.count(t) for t in range(8)]
      c_all = self.e_count = sum(c)
      e_start = self.e_type_start_index = [sum(c[:t]) for t in range(8)]
      e_slice = self.e_type_slice = [slice(e_start[t], e_start[t] + c[t]) for t in range(8)]
      # ****** create elements ids *************************
      self.e_ids = np.fromiter(map(lambda p: p[0], e_lines), np.int32)
      # ****** create elements nodes ***********************
      self.e_nodes = np.full((c_all, max(n for t, n in enumerate(self.E_type_nodes_count) if c[t] > 0)), None)
      self.e_nodes[0,0] = Node(0, (np.inf, np.inf, np.inf))
      id2node = lambda id: self.nodes[id-1]
      for t in range(1, 8):
         if c[t] > 0:
            for nodes, line in zip(self.e_nodes[e_slice[t], :self.E_type_nodes_count[t]], e_lines[e_slice[t]]):
               nodes[:] = np.fromiter(map(id2node, line[2:]), dtype="object")
      # ***********************************************
      for t, n in enumerate(c[1:], 1):
         if n > 0: self.info("Number of type {:n}: {:n}".format(t, n))
      self.info.OK("{:n} nodes, {:n} elements were loaded.".format(len(self.nodes), c_all - 1))


   def load_pflotran_ascii_regions(self, r_files, r_names):
      self.info.start("Loading PFloTran regions")
      id_e_dict = {id:i for i, id in enumerate(self.e_ids)}
      phs_id_start = 0 if len(self.physicals) == 0 else max(p.id for p in self.physicals.values()) + 1
      for phs_id, file_name in enumerate(r_files, phs_id_start):
         p = Physical(phs_id)
         p.dim = 3
         p.name = file_name if r_names is None else r_names[phs_id - phs_id_start]
         with open(file_name) as f:
            p.elements = np.fromiter(map(lambda s: id_e_dict[int(s.strip())], f.readlines()), np.int32)
         self.physicals[p.name] = p
         self.info("Region from file '" + file_name + "' was loaded.")
      self.info.OK()


   def load_pflotran_ascii_complete(self, ugi_file_name, r_dir, r_prefix, r_ext, br_dir, br_prefix, br_ext):
      self.info.start("Loading complete PFloTran model from ascii files")
      f, br =self.create_pflotran_regions_files_list(br_dir, br_prefix, self.pflotran_ascii_boundary_region_file_ext if br_ext is None else br_ext)
      self.load_pflotran_ascii_ugi_and_boundary_regions(ugi_file_name, f, br)
      f, r =self.create_pflotran_regions_files_list(r_dir, r_prefix, self.pflotran_ascii_regions_file_ext if r_ext is None else r_ext)
      self.load_pflotran_ascii_regions(f, r)
      self.info.OK()


   def save_pflotran_ascii_ugi(self, out_dir, file_name_without_ext):
      self.info.start("Saving PFLOTRAN unstructured grid to ascii file: '" + file_name_without_ext + "." + self.pflotran_ascii_ugi_file_ext + "'.")
      e4 = self.e_type_start_index[4]
      d = {n:i for i, n in enumerate(self.nodes, 1)}
      n2d = lambda n: d[n]
      nodes_count_2_line_format = {4:"T"+4*" {:n}"+"\n", 8:"H"+8*" {:n}"+"\n", 6:"W"+6*" {:n}"+"\n", 5:"P"+5*" {:n}"+"\n"}
      e4_nodes_count = np.count_nonzero(self.e_nodes[e4:], axis=1)
      ordered_e4 = self.e_ids[e4:].argsort()
      e4_nodes = self.e_nodes[e4:]
      with open(out_dir+"/"+file_name_without_ext + "." + self.pflotran_ascii_ugi_file_ext, "w") as f:
         f.write("{:n} {:n}\n".format(sum(self.e_type_count[4:]), len(self.nodes)))
         f.writelines(nodes_count_2_line_format[e4_nodes_count[e]].format(*map(n2d, e4_nodes[e,:e4_nodes_count[e]])) for e in ordered_e4)
         f.writelines(map(lambda n: "{:g} {:g} {:g}\n".format(n.x, n.y, n.z), self.nodes))
      self.info.OK()

      
   def save_pflotran_ascii_regions(self, out_dir, file_name_prefix):
      self.info.start("Saving NON boundary regions to ascii files")
      self.info("Directory: '" + out_dir + "'")
      self.info("File name prefix: '" + file_name_prefix + "'")
      for phs in self.physicals.values():
         if phs.dim != 3: continue
         with open(out_dir+"/"+file_name_prefix+phs.name+"."+self.pflotran_ascii_regions_file_ext, "w") as f:
            f.writelines(map(lambda e_id: "{:n}\n".format(e_id), np.sort(self.e_ids[phs.elements])))
      self.info.OK()


   def save_pflotran_ascii_boundary_regions(self, out_dir, file_name_prefix):
      self.info.start("Saving boundary regions to ascii files")
      self.info("Directory: '" + out_dir + "'")
      self.info("File name prefix: '" + file_name_prefix + "'")
      e2, e4 = self.e_type_start_index[2], self.e_type_start_index[4]
      d = {n:i for i, n in enumerate(self.nodes, 1)}
      n2d = lambda node: d[node]
      nodes_count_2_line_format = {3:"T"+3*" {:n}\n", 4:"Q"+4*" {:n}"+"\n"}
      e2_nodes_count = np.count_nonzero(self.e_nodes[e2:e4], axis=1)
      e2_nodes = self.e_nodes[e2:e4]
      for phs in self.physicals.values():
         if phs.dim != 2: continue
         ordered_e = phs.elements[self.e_ids[phs.elements].argsort()] - e2
         with open(out_dir+"/"+file_name_prefix+phs.name+"."+self.pflotran_ascii_boundary_region_file_ext, "w") as f:
            f.write("{:n}\n".format(phs.elements.size))
            f.writelines(map(lambda e: nodes_count_2_line_format[e2_nodes_count[e]].format(*map(n2d, e2_nodes[e,:e2_nodes_count[e]])) + "\n", ordered_e))
      self.info.OK()
   
   
   def save_pflotran_ascii_complete(self, out_dir, ugi_file_name_without_ext, r_prefix, br_prefix):
      self.info.start("Creating unstructured grid in ASCII format into directory '"+out_dir+"'.")
      self.save_pflotran_ascii_ugi(out_dir, ugi_file_name_without_ext)
      self.save_pflotran_ascii_regions(out_dir, r_prefix)
      self.save_pflotran_ascii_boundary_regions(out_dir, br_prefix)
      self.info.OK()


   # ********** PFLOTRAN hdf5 ***********************************

   def load_pflotran_h5_ugi_and_boundary_regions(self, ugi_group, reg_group):                                              
      self.info.start("Loading unstructured grid and boundary regions")
      e_id = itertools.count(1)
      # **** no element type *************
      e_lines = [(0, 0, 0)]
      # ***** loading mass elements and nodes *************************
      t_dict = {4:4, 5:7, 6:6, 8:5}
      e_lines.extend(map(lambda p: (next(e_id), t_dict[p[0]], *p[1:p[0]+1]), ugi_group["Cells"]))
      self.nodes = [Node(i, n) for i, n in enumerate(ugi_group["Vertices"])]
      self.info("Grid was loaded.")
      # *********** loading boundary regions ***************
      self.physicals = {}
      t_dict = {3:2, 4:3}
      phs_id = itertools.count(0)
      for name in reg_group:
         if "Vertex Ids" not in reg_group[name]: continue
         phs = Physical(next(phs_id))
         phs.name, phs.dim = name, 2
         phs.elements = [(next(e_id), t_dict[p[0]], *p[1:p[0]+1]) for p in reg_group[name]["Vertex Ids"]]
         e_lines.extend(phs.elements)
         self.physicals[phs.name] = phs
         self.info("Boundary region '" + name + "' was loaded.")
      # ***** sorting by element types *********************
      e_lines.sort(key=lambda s: s[1])
      # ****** set physicals elements by e_lines *******************
      e_line2e = {e_line:e for e, e_line in enumerate(e_lines)}
      for p in self.physicals.values():
         p.elements = np.fromiter(map(lambda e_line: e_line2e[e_line], p.elements), np.int32)
      # ****** slice by elements type ************************
      e_types = list(map(lambda p: p[1], e_lines))
      c = self.e_type_count = [e_types.count(t) for t in range(8)]
      c_all = self.e_count = sum(c)
      e_start = self.e_type_start_index = [sum(c[:t]) for t in range(8)]
      e_slice = self.e_type_slice = [slice(e_start[t], e_start[t] + c[t]) for t in range(8)]
      # ****** create elements ids *************************
      self.e_ids = np.fromiter(map(lambda p: p[0], e_lines), np.int32)
      # ****** create elements nodes ***********************
      self.e_nodes = np.full((c_all, max(n for t, n in enumerate(self.E_type_nodes_count) if c[t] > 0)), None)
      self.e_nodes[0,0] = Node(0, (np.inf, np.inf, np.inf))
      id2node = lambda id: self.nodes[id-1]
      for t in range(1, 8):
         if c[t] > 0:
            for nodes, line in zip(self.e_nodes[e_slice[t], :self.E_type_nodes_count[t]], e_lines[e_slice[t]]):
               nodes[:] = np.fromiter(map(id2node, line[2:]), dtype="object")
      # ***********************************************
      for t, n in enumerate(c[1:], 1):
         if n > 0: self.info("Number of type {:n}: {:n}".format(t, n))
      self.info.OK("{:n} nodes, {:n} elements were loaded.".format(len(self.nodes), c_all - 1))


   def load_pflotran_h5_regions(self, group):
      self.info.start("Loading regions")
      phs_id = itertools.count(0 if len(self.physicals) == 0 else max(p.id for p in self.physicals.values()) + 1)
      for name, group in group.items():
         if "Cell Ids" not in group: continue
         p = Physical(next(phs_id))
         p.dim, p.name = 3, name
         p.elements = np.fromiter(group["Cell Ids"], np.int32)
         self.physicals[name] = p
         self.info("Region '" + name + "' was loaded.")
      self.info.OK()


   def load_pflotran_h5_complete(self, file_name):
      self.info.start("Loading Pflotran complete model from h5 file '" + file_name + "'.")
      with h5.File(file_name, mode="r") as f:
         ugi = f["Domain"]
         r = f["Regions"]
         self.load_pflotran_h5_ugi_and_boundary_regions(ugi, r)
         self.load_pflotran_h5_regions(r)
      self.info.OK()


   def save_pflotran_h5_ugi(self, group):
      self.info.start("Creating Cells dataset")
      e4 = self.e_type_start_index[4]
      cells = np.zeros((self.e_count - e4, 9), np.int32)
      cells[:,0] = np.count_nonzero(self.e_nodes[e4:], axis=1)
      d = {n:i for i, n in enumerate(self.nodes, 1)} | {None:0}
      n2d = np.vectorize(lambda n: d[n], [np.int32])      
      cells[:,1:self.e_nodes.shape[1]+1] = n2d(self.e_nodes[e4:])
      group.create_dataset("Cells", data=cells[self.e_ids[e4:].argsort()])
      vertices = np.array([(n.x, n.y, n.z) for n in self.nodes], np.float64)
      group.create_dataset("Vertices", data=vertices)
      self.info.OK()


   def save_pflotran_h5_regions(self, group):
      self.info.start("Creating regions")
      for phs in self.physicals.values():
         if phs.dim != 3: continue
         g = group.create_group(phs.name)
         g.create_dataset("Cell Ids", data=np.sort(self.e_ids[phs.elements]), dtype=np.int32)
      self.info.OK()


   def save_pflotran_h5_boundary_regions(self, group):
      self.info.start("Creating boundary regions")
      d = {n:i for i, n in enumerate(self.nodes, 1)} | {None:0}
      n2d = np.vectorize(lambda n: d[n], [np.int32])
      for phs in self.physicals.values():
         if phs.dim != 2: continue
         bc = np.zeros((phs.elements.size, 5), np.int32)
         bc[:,0] = np.count_nonzero(self.e_nodes[phs.elements], axis=1)
         bc[:,1:] = n2d(self.e_nodes[phs.elements,:4])
         g = group.create_group(phs.name)
         g.create_dataset("Vertex Ids", data=bc[self.e_ids[phs.elements].argsort()], dtype=np.int32)
      self.info.OK()


   def save_pflotran_h5_complete(self, file_name):
      self.info.start("Creating unstructured grid in h5 format in file '"+file_name+"'.")
      with h5.File(file_name, "w") as f:
         group = f.create_group("Domain")
         self.save_pflotran_h5_ugi(group)
         group_regions = f.create_group("Regions")
         self.save_pflotran_h5_regions(group_regions)
         self.save_pflotran_h5_boundary_regions(group_regions)
      self.info.OK()


   # ********** base analytical tools ********************

   def get_e_count(self):
      return len(self.e_nodes)

   def get_e_type_count(self, type):
      return (self.e_type_start_index[type+1] if type < 7 else len(self.e_nodes)) - self.e_type_start_index[type]

   def get_e_type_slice(self, type):
      return slice(self.e_type_start_index[type], self.e_type_start_index[type+1] if type < 7 else len(self.e_nodes) )

   def choose_e_type(self, type, elems, bool_indices=False):
      if type < 7: B = (elems >= self.e_type_start_index[type]) & (elems < self.e_type_start_index[type + 1])
      else: B = elems >= self.e_type_start_index[7]
      if bool_indices: return B   
      else: return elems[B]
                                                              
   def get_e_type_nodes(self, type):
      return self.e_nodes[self.get_e_type_slice(type), :self.E_type_nodes_count[type]]

   def get_e_type_nodes_xyz(self, type):
      return self.nodes_xyz[self.e_nodes[self.get_e_type_slice(type), :self.E_type_nodes_count[type]]]

   def get_e_type_nodes_x(self, type):
      return self.nodes_xyz[self.e_nodes[self.get_e_type_slice(type), :self.E_type_nodes_count[type]], 0]

   def get_e_type_nodes_y(self, type):
      return self.nodes_xyz[self.e_nodes[self.get_e_type_slice(type), :self.E_type_nodes_count[type]], 1]

   def get_e_type_nodes_z(self, type):
      return self.nodes_xyz[self.e_nodes[self.get_e_type_slice(type), :self.E_type_nodes_count[type]], 2]



   # *********** addition analytical tools **************

   def get_e_physicals_ids(self):
      e_phs = np.full(self.get_e_count(), -1, np.int32)
      for p in self.physicals.values(): e_phs[p.elements] = p.id         
      return e_phs
      

   def get_nodes_of_physicals(self, physicals_ids, info=False):
      if info: self.info.start("Getting nodes of selected physicals")
      nodes = np.unique(np.concatenate([(n:=self.e_nodes[self.physicals[p_id].elements].reshape(-1))[n!=NO_NODE_INDEX] for p_id in physicals_ids]))
      nodes.sort()
      if info:
         self.info(str(len(nodes))+" nodes were found.")
         self.info.OK()
      return nodes

   def get_elements_of_physicals(self, physicals_ids, info=False):
      if info: self.info.start("Getting elements of selected physicals")
      elems = np.concatenate([self.physicals[id].elements for id in physicals_ids])
      elems.sort()
      if info:
         self.info(str(len(elems))+" elements were found.")
         self.info.OK()
      return elems
   

   


# ******* Mesh for ToolKit ***************************

class MeshForToolKits(Mesh):

   def __init__(self, info):
      super().__init__(info)
      self.above_toolkits = []

   def update_above_toolkits(self):
      for t in self.above_toolkits: t.update()


