import numpy as np

from .mesh_toolkit import MeshToolKit

class ParamsToolKit(MeshToolKit):

   class Reporting_error(Exception): pass

   def __init__(self, mesh):
      super().__init__(mesh=mesh)
      self.OPTIONS_keep_e_T = True
      self.OPTIONS_keep_e_ranges = False
      self.OPTIONS_keep_e_planes_params = False
      self.OPTIONS_keep_tria_params = False
      self.OPTIONS_keep_nodes_e = False
      self.OPTIONS_keep_e4_neighbours_e2e4_by_plane = False
      self.OPTIONS_keep_e2_neighbours_e4_by_plane = False
      self.OPTIONS_keep_e_neighbours = False
      self.OPTIONS_top_boundary_physicals_ids = []
      self.x_ref, self.y_ref = 0., 0.


   def update_method(self):
      self.x_0, self.y_0, self.z_0 = np.min(self.mesh.nodes_xyz, axis=0)
      self.x_max, self.y_max, self.z_max = np.max(self.mesh.nodes_xyz, axis=0)
      self.x_range, self.y_range, self.z_range = self.x_max - self.x_0, self.y_max - self.y_0, self.z_max - self.z_0
      self.info("X axis (minimum, maximum, range): {:g}\t{:g}\t{:g}".format(self.x_0, self.x_max, self.x_range))
      self.info("Y axis (minimum, maximum, range): {:g}\t{:g}\t{:g}".format(self.y_0, self.y_max, self.y_range))
      self.info("Z axis (minimum, maximum, range): {:g}\t{:g}\t{:g}".format(self.z_0, self.z_max, self.z_range))
      self.e_T = None
      self.e_min_x = self.e_min_y = self.e_min_z = None
      self.e_max_x = self.e_max_y = self.e_max_z = None
      self.e_planes_abcd = 8 * [None]
      self.tria_nodes = None
      self.tria_xy_sides_ab_c  = None
      self.tria_planes_abc_d = None
      self.nodes_e = self._old_nodes_e_iter_types = None      
      self.e4_neighbours_e2e4_by_plane = None
      self.e2_neighbours_e4_by_plane = None
      self.e_neighbours = self._old_e_neighbours_by_nodes_num = None
      self.OPTIONS_top_boundary_physicals_ids = []
      self.OPTIONS_physicals_cross_sections = {}
      self.OPTIONS_default_physicals_cross_section = 1

      
   # ********* calc by reference point ************
   
   def get_xy_by_ref(self, ofset_x, ofset_y):
      x = None if ofset_x is None else self.x_ref + ofset_x
      y = None if ofset_y is None else self.y_ref + ofset_y
      return x, y


   # ********* elements T **************************

   def get_e_T(self):
      if self.OPTIONS_keep_e_T:
         if self.e_T is None: self.e_T = self.calc_e_T()
         return self.e_T
      else: return self.calc_e_T()

   def calc_e_T(self, mesh=None):
      self.info.start("Calculating centers of elements")
      m = self.mesh if mesh is None else mesh
      T = np.empty((m.get_e_count(), 3), np.float64)    
      for t in range(8):
         if m.get_e_type_count(t) > 0:
            T[m.get_e_type_slice(t)] = np.mean(m.get_e_type_nodes_xyz(t), axis=1)
      self.info.OK()
      return T


   # ******* elements min, max ******************

   def get_e_min(self, osa):
      e_min = (self.e_min_x, self.e_min_y, self.e_min_z)[osa]
      if self.OPTIONS_keep_e_ranges:
         if e_min is None: e_min = self.calc_e_min(osa)
         return e_min
      else: return self.calc_e_min(osa)

   def calc_e_min(self, osa):
      self.info.start("Calculating minimum "+("x","y","z")[osa]+" coordinate of elements")
      m = self.mesh
      Min = np.empty(m.get_e_count(), np.float64)    
      for t in range(8):
         if m.get_e_type_count(t) > 0:
            s, c = m.get_e_type_slice(t), m.E_type_nodes_count[t]
            Min[s] = np.min(m.nodes_xyz[m.e_nodes[s,:c],osa], axis=1)
      self.info.OK()
      return Min


   def get_e_max(self, osa):
      e_max = (self.e_max_x, self.e_max_y, self.e_max_z)[osa]
      if self.OPTIONS_keep_e_ranges:
         if e_max is None: e_max = self.calc_e_max(osa)
         return e_max
      else: return self.calc_e_max(osa)

   def calc_e_max(self, osa):
      self.info.start("Calculating maximum "+("x","y","z")[osa]+" coordinate of elements")
      m = self.mesh
      Max = np.empty(m.get_e_count(), np.float64)    
      for t in range(8):
         if m.get_e_type_count(t) > 0:
            s, c = m.get_e_type_slice(t), m.E_type_nodes_count[t]
            Max[s] = np.max(m.nodes_xyz[m.e_nodes[s,:c],osa], axis=1)
      self.info.OK()
      return Max


   # ******* elements planes ********************

   def get_e_type_planes_abcd(self, t):
      if self.OPTIONS_keep_e_planes_params:
         if self.e_planes_abcd[t] is None: self.e_planes_params[t] = eval("self.calc_e{:n}_planes_abcd()".format(t))
         return self.e_planes_abcd[t]
      else: 
         return eval("self.calc_e{:n}_planes_abcd()".format(t))
         

   def calc_e2_planes_abcd(self):
      self.info.start("Calculating planes params for 2D elemennts (type 2 - triangles)")
      p = self.mesh.get_e_type_nodes_xyz(2)
      abcd = np.empty((len(p), 4, 4), np.float64)
      abcd[:,0,:3] = n = np.cross(p[:,1]-p[:,0], p[:,2]-p[:,0])
      abcd[:,1,:3] = np.cross(n, p[:,1] - p[:,0])
      abcd[:,2,:3] = np.cross(n, p[:,2] - p[:,1])
      abcd[:,3,:3] = np.cross(n, p[:,0] - p[:,2])
      abcd[:,:,:3] /= (np.sum(abcd[:,:,:3]**2, axis=2)**0.5)[:,:,None]    
      abcd[:,0,3] = - np.sum(abcd[:,0,:3] * p[:,0], axis=1)
      abcd[:,1,3] = - np.sum(abcd[:,1,:3] * p[:,0], axis=1)
      abcd[:,2,3] = - np.sum(abcd[:,2,:3] * p[:,1], axis=1)
      abcd[:,3,3] = - np.sum(abcd[:,3,:3] * p[:,2], axis=1)
      self.info.OK()
      return abcd

   def calc_e3_planes_abcd(self):
      self.info.start("Calculating planes params for 2D elemennts (type 3 - quad)")
      p = self.mesh.get_e_type_nodes_xyz(3)
      abcd = np.empty((len(p), 1, 4), np.float64)
      abcd[:,0,:3] = np.cross(p[:,1]-p[:,0], p[:,3]-p[:,0])
      abcd[:,:,:3] /= (np.sum(abcd[:,:,:3]**2, axis=1)**0.5)[:,:,None]
      abcd[:,0,3] = - np.sum(abcd[:,:,:3] * p[:,0], axis=1)
      self.info.OK()
      return abcd

   def calc_e4_planes_abcd(self):
      self.info.start("Calculating planes params for 3D elements type 4 (tetrahedron)")
      p = self.mesh.get_e_type_nodes_xyz(4)
      abcd = np.empty((self.mesh.get_e_type_count(4), 4, 4), np.float64)
      abcd[:,0,:3] = np.cross(p[:,3]-p[:,0], p[:,1]-p[:,0])
      abcd[:,1,:3] = np.cross(p[:,3]-p[:,1], p[:,2]-p[:,1])
      abcd[:,2,:3] = np.cross(p[:,3]-p[:,2], p[:,0]-p[:,2])
      abcd[:,3,:3] = np.cross(p[:,1]-p[:,0], p[:,2]-p[:,0])
      abcd[:,:,:3] /= (np.sum(abcd[:,:,:3]**2, axis=2)**0.5)[:,:,None]      
      abcd[:,0,3] = - np.sum(abcd[:,0,:3] * p[:,0], axis=1)
      abcd[:,1,3] = - np.sum(abcd[:,1,:3] * p[:,1], axis=1)
      abcd[:,2,3] = - np.sum(abcd[:,2,:3] * p[:,2], axis=1)
      abcd[:,3,3] = - np.sum(abcd[:,3,:3] * p[:,0], axis=1)
      self.info.OK()
      return abcd

   def calc_e5_planes_abc_d(self):
      self.info.start("Calculating planes params for 3D elements type 5 (hexahedron)")
      p = self.mesh.get_e_type_nodes_xyz(5)
      abc = np.empty((self.mesh.get_e_type_count(5), 6, 3), np.float64)
      # ********** plast **********************
      abc[:,0] = np.cross(p[:,4]-p[:,0], p[:,1]-p[:,0])
      abc[:,1] = np.cross(p[:,5]-p[:,1], p[:,2]-p[:,1])
      abc[:,2] = np.cross(p[:,6]-p[:,2], p[:,3]-p[:,2])
      abc[:,3] = np.cross(p[:,7]-p[:,3], p[:,0]-p[:,3])
      # ********** podstava **********************
      abc[:,4] = np.cross(p[:,1]-p[:,0], p[:,3]-p[:,0])
      abc[:,5] = np.cross(p[:,7]-p[:,4], p[:,5]-p[:,4])
      # *********** normalizace ************************
      abc /= (np.sum(abc**2, axis=2)**0.5)[:,:,None]
      # ************************************************
      d = np.empty((self.mesh.get_e_type_count(5), 6), np.float64)
      d[:,0] = - np.sum(abc[:,0] * p[:,0], axis=1)
      d[:,1] = - np.sum(abc[:,1] * p[:,1], axis=1)
      d[:,2] = - np.sum(abc[:,2] * p[:,2], axis=1)
      d[:,3] = - np.sum(abc[:,3] * p[:,3], axis=1)
      d[:,4] = - np.sum(abc[:,4] * p[:,0], axis=1)
      d[:,5] = - np.sum(abc[:,5] * p[:,4], axis=1)
      self.info.OK()
      return abc, d

   def calc_e6_planes_abc_d(self):
      self.info.start("Calculating planes params for 3D elements type 6 (wedge)")
      p = self.mesh.get_e_type_nodes_xyz(6)
      abc = np.empty((self.mesh.get_e_type_count(6), 5, 3), np.float64)
      # ********** plast **********************
      abc[:,0] = np.cross(p[:,3]-p[:,0], p[:,1]-p[:,0])
      abc[:,1] = np.cross(p[:,4]-p[:,1], p[:,2]-p[:,1])
      abc[:,2] = np.cross(p[:,5]-p[:,2], p[:,0]-p[:,2])
      # ********** podstava **********************
      abc[:,4] = np.cross(p[:,1]-p[:,0], p[:,2]-p[:,0])
      abc[:,5] = np.cross(p[:,5]-p[:,3], p[:,4]-p[:,3])
      # ************** normalizace *********************
      abc /= (np.sum(abc**2, axis=2)**0.5)[:,:,None]
      # ************************************************
      d = np.empty((self.mesh.get_e_type_count(6), 5), np.float64)
      d[:,0] = - np.sum(abc[:,0] * p[:,0], axis=1)
      d[:,1] = - np.sum(abc[:,1] * p[:,1], axis=1)
      d[:,2] = - np.sum(abc[:,2] * p[:,2], axis=1)
      d[:,3] = - np.sum(abc[:,3] * p[:,0], axis=1)
      d[:,4] = - np.sum(abc[:,4] * p[:,3], axis=1)
      self.info.OK()
      return abc, d

   def calc_e7_planes_abc_d(self):
      self.info.start("Calculating planes params for 3D elements type 7 (pyramid)")
      p = self.mesh.get_e_type_nodes_xyz(7)
      abc = np.empty((self.mesh.get_e_type_count(7), 5, 3), np.float64)
      # ********** plast **********************
      abc[:,0] = np.cross(p[:,4]-p[:,0], p[:,1]-p[:,0])
      abc[:,1] = np.cross(p[:,4]-p[:,1], p[:,2]-p[:,1])
      abc[:,2] = np.cross(p[:,4]-p[:,2], p[:,3]-p[:,2])
      abc[:,3] = np.cross(p[:,4]-p[:,3], p[:,0]-p[:,3])
      # ********** podstava **********************
      abc[:,4] = np.cross(p[:,1]-p[:,0], p[:,3]-p[:,0])
      # ************* normalizace **********************
      abc /= (np.sum(abc**2, axis=2)**0.5)[:,:,None]
      # ************************************************
      d = np.empty((self.mesh.get_e_type_count(7), 5), np.float64)
      d[:,0] = - np.sum(abc[:,0] * p[:,0], axis=1)
      d[:,1] = - np.sum(abc[:,1] * p[:,1], axis=1)
      d[:,2] = - np.sum(abc[:,2] * p[:,2], axis=1)
      d[:,3] = - np.sum(abc[:,3] * p[:,3], axis=1)
      d[:,4] = - np.sum(abc[:,4] * p[:,0], axis=1)
      self.info.OK()
      return abc, d


   # ************ elements volumes ********************
   
   def calc_e_volumes(self, e):
      self.info.start("Calculating volumes of {:n} elements".format(len(e)))
      m = self.mesh
      if np.any(e >= m.e_type_start_index[5]):
         raise self.ToolKitException("Volume can be calculated for elements types 1, 2, 3, 4")            
      e_V = np.full(len(e), np.nan, dtype=np.float64)
      if np.any(B:=m.choose_e_type(1, e, bool_indices=True)):
         e1 = e[B] - m.e_type_start_index[1]
         r = self.get_e_type_cross_sections(1)[e1]
         n = m.get_e_type_nodes_xyz(1)[e1]
         e_V[B] = r**2 * np.sum((n[:,1] - n[:,0])**2, axis=1)**0.5         
      if np.any(B:=m.choose_e_type(2, e, bool_indices=True)):
         e2 = e[B] - m.e_type_start_index[2]
         r = self.get_e_type_cross_sections(2)[e2]
         n = m.get_e_type_nodes_xyz(2)[e2]
         e_V[B] = r * np.sum(np.cross(n[:,1] - n[:,0], n[:,2] - n[:,0])**2, axis=1)**0.5 / 2
      if np.any(B:=m.choose_e_type(3, e, bool_indices=True)):
         e3 = e[B] - m.e_type_start_index[3]
         r = self.get_e_type_cross_sections(3)[e3]
         n = m.get_e_type_nodes_xyz(3)[e3]
         e_V[B] = r / 2 * ( \
              np.sum(np.cross(n[:,1] - n[:,0], n[:,2] - n[:,0])**2, axis=1)**0.5 \
            + np.sum(np.cross(n[:,3] - n[:,0], n[:,2] - n[:,0])**2, axis=1)**0.5 )
      if np.any(B:=m.choose_e_type(4, e, bool_indices=True)):
         e4 = e[B] - m.e_type_start_index[4]
         n = m.get_e_type_nodes_xyz(4)[e4]
         e_V[B] = np.absolute(np.linalg.det(n[:,:3] - n[:,3,None])) / 6
      self.info.OK()
      return e_V


   # ***************** elements areas ******************

   def calc_e_areas(self, e):
      self.info.start("Calculating areas of {:n} elements types 2, 3".format(len(e)))
      m = self.mesh
      if np.any(np.logical_or(e < m.e_type_start_index[2], e >= m.e_type_start_index[4])): 
         raise self.ToolKitException("Area can be calculated only for elements types 2, 3")            
      e_S = np.full(len(e), np.nan, dtype=np.float64)
      if np.any(B:=m.choose_e_type(2, e, bool_indices=True)):
         e2 = e[B] - m.e_type_start_index[2]
         n = m.get_e_type_nodes_xyz(2)[e2]
         e_S[B] = np.sum(np.cross(n[:,1] - n[:,0], n[:,2] - n[:,0])**2, axis=1)**0.5 / 2
      if np.any(B:=m.choose_e_type(3, e, bool_indices=True)):
         e3 = e[B] - m.e_type_start_index[3]
         n = m.get_e_type_nodes_xyz(3)[e3]
         e_S[B] = 1 / 2 * ( \
              np.sum(np.cross(n[:,1] - n[:,0], n[:,2] - n[:,0])**2, axis=1)**0.5 \
            + np.sum(np.cross(n[:,3] - n[:,0], n[:,2] - n[:,0])**2, axis=1)**0.5 )
      self.info.OK()
      return e_S


   # ************ triangles *****************************

   def create_tria_nodes(self):
      self.info.start("Create triangles")
      if len(self.OPTIONS_top_boundary_physicals_ids) == 0:
         raise self.ToolKitException("No top boundary entity is set.")
      elems_2 = np.concatenate([self.mesh.physicals[p_id].elements for p_id in self.OPTIONS_top_boundary_physicals_ids])
      self.tria_nodes = self.mesh.e_nodes[elems_2, :3]
      self.info.OK()


   def get_tria_xy_sides_ab_c(self):
      if self.OPTIONS_keep_tria_params:
         if self.tria_xy_sides_ab_c is None: self.tria_xy_sides_ab_c = self.calc_tria_xy_sides_ab_c()
         return self.tria_xy_sides_ab_c
      else: return self.calc_tria_xy_sides_ab_c()


   def calc_tria_xy_sides_ab_c(self):
      self.info.start("Calculating xy sides of triangles")
      n_xy = self.mesh.nodes_xyz[self.tria_nodes,:2]
      x, y = n_xy[:,:,0], n_xy[:,:,1]
      ab = np.empty((len(n_xy), 3, 2), np.float64)
      ab[:,0,0], ab[:,0,1] = - (y[:,1] - y[:,0]), x[:,1] - x[:,0]
      ab[:,1,0], ab[:,1,1] = - (y[:,2] - y[:,1]), x[:,2] - x[:,1]
      ab[:,2,0], ab[:,2,1] = - (y[:,0] - y[:,2]), x[:,0] - x[:,2]
      ab /= (np.sum(ab**2, axis=2)**0.5)[:,:,None]
      c = - np.sum(ab * n_xy, axis=2, dtype=np.float64)
      B = np.sum(ab[:,0] * n_xy[:,2], axis=1) < - c[:,0]
      ab[B,0] *= -1; c[B,0] *= -1
      B = np.sum(ab[:,1] * n_xy[:,0], axis=1) < - c[:,1]
      ab[B,1] *= -1; c[B,1] *= -1
      B = np.sum(ab[:,2] * n_xy[:,1], axis=1) < - c[:,2]
      ab[B,2] *= -1; c[B,2] *= -1
      self.info.OK()
      return ab, c


   def get_tria_planes_abc_d(self):
      if self.OPTIONS_keep_tria_params:
         if self.tria_planes_abc_d is None: self.tria_planes_abc_d = self.calc_tria_planes_abc_d()
         return self.tria_planes_abc_d
      else: return self.calc_tria_planes_abc_d()


   def calc_tria_planes_abc_d(self):
      self.info.start("Calculating planes of triangles")
      p = self.mesh.nodes_xyz[self.tria_nodes]
      abc = np.cross(p[:,1]-p[:,0], p[:,2]-p[:,0])
      abc /= (np.sum(abc**2, axis=1)**0.5)[:,None]
      d = - np.sum(abc * p[:,0], axis=1)
      self.info.OK()
      return abc, d


   # ******** Neighboring *************************
   
   def get_nodes_elements(self, iter_types=list, e_types=[0,1,2,3,4,5,6,7]):
      if self.OPTIONS_keep_nodes_e:
         if self._old_nodes_e_iter_types != (iter_types, e_types):
            self.nodes_e = self.calc_nodes_elements(iter_types=iter_types, e_types=e_types)
            self._old_nodes_e_iter_type = (iter_types, e_types)
         return self.nodes_e
      else: return self.calc_nodes_elements(iter_types=iter_types, e_types=e_types)
   

   def calc_nodes_elements(self, iter_types=list, e_types=[0,1,2,3,4,5,6,7]):
      self.info.start("Searching elements of nodes")
      m = self.mesh
      nodes_e = np.fromiter([[] for _ in range(len(m.nodes_id))], object)
      for t in e_types:
         if m.get_e_type_count(t) > 0:
            sc, nc = m.get_e_type_slice(t), m.E_type_nodes_count[t]
            for e, nodes in enumerate(m.e_nodes[sc,:nc], m.e_type_start_index[t]):
               for n in nodes: nodes_e[n].append(e)
      if iter_types == tuple:
         for i in range(len(nodes_e)): nodes_e[i] = tuple(nodes_e[i])               
      if iter_types == set:
         for i in range(len(nodes_e)): nodes_e[i] = set(nodes_e[i])
      elif iter_types == np.array:
         for i in range(len(nodes_e)): nodes_e[i] = np.array(nodes_e[i], np.int32)
      self.info.OK()
      return nodes_e


   def get_e4_neighbours_e2e4_by_plane(self, e4_selection=None):
      if self.OPTIONS_keep_e4_neighbours_e2e4_by_plane:
         if self.e4_neighbours_e2e4_by_plane is None: self.e4_neighbours_e2e4_by_plane = self.calc_e4_neighbours_e2e4_by_plane()
         if e4_selection is None:
            return self.e4_neighbours_e2e4_by_plane
         else: return self.e4_neighbours_e2e4_by_plane[e4_selection]
      else: return self.calc_e4_neighbours_e2e4_by_plane(e4_selection=e4_selection)

   def calc_e4_neighbours_e2e4_by_plane(self, e4_selection=None):
      self.info.start("Searching neighbours of elements type 4 by surface")
      m = self.mesh
      if e4_selection is None:
         sc = m.get_e_type_slice(4)
         elems = np.arange(sc.start, sc.stop, dtype=np.int32)
      else: elems = e4_selection
      e_s_n = m.e_nodes[elems][:,[[0,1,2],[0,1,3],[0,2,3],[1,2,3]]]
      n_e = self.get_nodes_elements(iter_types=set, e_types=[2,4])
      e_s_n_Es = n_e[e_s_n]      
      e_s_Enb = np.full((len(elems), 4), -1, np.int32)
      for e, s_Enb, s_n_Es in zip(elems, e_s_Enb, e_s_n_Es):
         for i, n_Es in enumerate(s_n_Es):
            b = n_Es[0] & n_Es[1] & n_Es[2]
            b.remove(e)
            if len(b) == 1: s_Enb[i] = b.pop()
            elif len(b) == 2: s_Enb[i] = min(b)
      self.info.OK()
      return e_s_Enb


   def get_e1e2_neighbours_of_e2_by_edge(self):
      pass


   def get_e2_neighbours_e4_by_plane(self, e2_selection=None):
      if self.OPTIONS_keep_e2_neighbours_e4_by_plane:
         if self.e2_neighbours_e4_by_plane is None: self.e2_neighbours_e4_by_plane = self.calc_e2_neighbours_e4_by_plane()
         if e2_selection is None:
            return self.e2_neighbours_e4_by_plane
         else: return self.e2_neighbours_e4_by_plane[e2_selection]
      else: return self.calc_e2_neighbours_e4_by_plane(e2_selection=e2_selection)

   def calc_e2_neighbours_e4_by_plane(self, e2_selection):
      self.info.start("Searching neighbours of elements with type 2 by surface")
      m = self.mesh
      if e2_selection is None:
         sc = m.get_e_type_slice(2)
         elems = np.arange(sc.start, sc.stop, dtype=np.int32)
      else: elems = e2_selection
      n_e = self.get_nodes_elements(iter_types=set, e_types=[4])
      e_n_Es = n_e[m.e_nodes[elems,:3]]
      e_Enb = np.full((len(elems), 2), -1, np.int32)      
      for Enb, n_Es in zip(e_Enb, e_n_Es):
         b = n_Es[0] & n_Es[1] & n_Es[2]
         if len(b) > 1: Enb[:] = list(b)
         elif len(b) == 1: Enb[0] = b.pop()
      self.info.OK()
      return e_Enb      







   def get_e_neighbours(self, by_nodes_num, e_selection=None):
      if self.OPTIONS_keep_e_neighbours:
         if self._old_e_neighbours_by_nodes_num != by_nodes_num:
            self.e_neighbours = self.calc_e_neighbours(by_nodes_num)
            self._old_e_neighbours_by_nodes_num = by_nodes_num
         if e_selection is None: return self.e_neighbours
         else: return self.e_neighbours[e_selection]
      else: return self.calc_e_neighbours(by_nodes_num, e_selection=e_selection)


   def calc_e_neighbours(self, by_nodes_num, e_selection):
      self.info.start("Searching elements neighbours")      
      m = self.mesh
      elems = np.arange(0, m.get_e_count(), dtype=np.int32) if e_selection is None else e_selection      
      n_e = self.get_nodes_elements(iter_types=tuple)
      R = {}
      for t, c in enumerate(m.E_type_nodes_count):
         ets = m.choose_e_type(t, elems)
         if len(ets) == 0: continue 
         e_n_Es = n_e[m.e_nodes[ets,:c]]
         if by_nodes_num is None:
            for e, n_Es in zip(ets, e_n_Es):   
               b = set(n_Es[0])
               for Es in n_Es[1:]: b.update(Es)
               b.remove(e)               
               nbs = np.fromiter([e] + list(b), np.int32)
               R.setdefault(len(nbs)-1, []).append(nbs[None,:])
         else:
            for e, n_Es in zip(ets, e_n_Es):   
               b = {s:1 for s in n_Es[0]}   
               for Es in n_Es[1:]:
                  for s in Es:
                     if s in b: b[s] += 1
                     else: b[s] = 1
               b.pop(e)
               nbs = np.fromiter([e] + [s for s in b if b[s] in by_nodes_num], np.int32)
               R.setdefault(len(nbs)-1, []).append(nbs[None,:])
      for c in R: R[c] = np.concatenate(R[c], dtype=np.int32)
      self.info.OK()
      return R


   def get_e_neighbours_num(self, by_nodes_num, e_selection=None):
      self.info.start("Calculating elements neighbours")
      B = self.get_e_neighbours(by_nodes_num, e_selection=e_selection)
      r = sorted(((c, len(B[c])) for c in B))
      for c, ec in r: print(c, ec)         
      self.info.OK()
      return r


   # *********** top boundary physicals ************

   def get_top_boundary_physicals_selection(self):
      self.info.start("Getting top boundary physicals")
      names = {p.id:(p.name, p.dim) for p in self.mesh.physicals.values()}
      selected = list(self.OPTIONS_top_boundary_physicals_ids)
      self.info.OK()
      return names, selected


   def set_top_boundary_physicals_selection(self, physicals_ids):
      self.info.start("Setting top boundary physicals")
      self.OPTIONS_top_boundary_physicals_ids = list(physicals_ids)
      self.top_boundary_triangles = None
      self.info.OK()


   def get_e_type_cross_sections(self, t):
      self.info.start("Setting cross-sections of elements type 2 by physicals")
      m = self.mesh      
      e_r = np.full(m.get_e_type_count(t), self.OPTIONS_default_physicals_cross_section, dtype=np.float64)
      for id, w in self.OPTIONS_physicals_cross_sections.items():
         e_r[m.choose_e_type(t, m.physicals[id].elements) - m.e_type_start_index[t]] = w
      self.info.OK()
      return e_r




