import numpy as np
import h5py as h5

from .mesh_toolkit import MeshToolKit

class ReportToolKit(MeshToolKit):

   class Reporting_error(Exception): pass

   def __init__(self, params, DEM, bands, data):
      super().__init__(params=params, DEM=DEM, bands=bands, data=data)


   def report_model_range(self):
      self.info.start("Getting global mesh ranges")
      p = self.params
      dialog = [(p.x_0, p.x_max, p.x_range),
                (p.y_0, p.y_max, p.y_range),
                (p.z_0, p.z_max, p.z_range) ]      
      protocol = [("x min.", p.x_0), ("x max.", p.x_max), ("x range", p.x_range),
                  ("y min.", p.y_0), ("y max.", p.y_max), ("y range", p.y_range),
                  ("z min.", p.z_0), ("z max.", p.z_max), ("z range", p.z_range) ]
      self.info.OK()
      return dialog, protocol


   def report_DEM_range_of_mesh_area(self):
      self.info.start("Reporting DEM range in mesh area")
      R = self.params
      p = self.DEM.get_xyz_points(x0=R.x_0, y0=R.y_0, x_max=R.x_max, y_max=R.y_max)
      b = self.bands.is_in_surf(p[:,:2])
      min_z, max_z = p[b,2].min(), p[b,2].max()
      dialog = [min_z, max_z, max_z-min_z]
      protocol = [("Minimum DEM elevation of model area", min_z), ("Maximu DEM elemation of model area", max_z), ("Range DEM elevation of model area", max_z-min_z)]
      self.info.OK()
      return dialog, protocol


   def report_physicals_group(self, ids):
      self.info.start("Reporting physicals group with physicals ids " + (len(ids)*" {:n}").format(*ids))
      D = {self.mesh.physicals[id].dim for id in ids}
      m = self.mesh
      elms = m.get_elements_of_physicals(ids)
      if D == {2}: S = np.sum(self.params.calc_e_areas(elms))
      V = np.sum(self.params.calc_e_volumes(elms))
      n = m.nodes_xyz[m.get_nodes_of_physicals(ids)]
      x0, y0, z0 = np.min(n, axis=0)
      x1, y1, z1 = np.max(n, axis=0)
      dialog = [(x0, x1, x1-x0), (y0, y1, y1-y0), (z0, z1, z1-z0), 
                L if D == {1} else None, S if D == {2} else None, V]
      protocol = [("x min.", x0), ("x max.", x1), ("x range", x1-x0),
                  ("y min.", y0), ("y max.", y1), ("y range", y1-y0),
                  ("z min.", z0), ("z max.", z1), ("z range", z1-z0),
                  ("Length", L if D == {1} else None),
                  ("Area", S if D =={2} else None),
                  ("Volume", V) ]
      self.info.OK()
      return dialog, protocol


   def calc_e4_flux_over_e2e4_neighbours(self, e4_physicals_ids, e2e4_neighbours_physicals_ids, T, vector):
      self.info.start("Calculating flux of physicals with elements of type 4 over neighbouring physicals with elements with type 2 or 4.")
      a = [id for id in e4_physicals_ids if self.mesh.physicals[id].dim != 3]
      if len(a) > 0: raise self.ToolKitException("Physicals " + ", ".join([str(id) for id in a]) + " are not 3D with elements type 4.")
      e4 = self.mesh.get_elements_of_physicals(e4_physicals_ids)
      a = [id for id in e2e4_neighbours_physicals_ids if self.mesh.physicals[id].dim == 1]
      if len(a) > 0: raise self.ToolKitException("Physicals " + ", ".join([str(id) for id in a]) + " are not 2D or 3D with elements type 2 or 4.")
      e2e4_nbs = self.mesh.get_elements_of_physicals(e2e4_neighbours_physicals_ids)      
      f_nb = self.params.get_e4_neighbours_e2e4_by_plane(e4_selection=e4)
      B = np.zeros(f_nb.shape, np.bool)      
      for i in range(4): B[:,i] = np.in1d(f_nb[:,i], e2e4_nbs)
      Q = self.data.raw_ep_ef_sp_sf[T][3][e4][B]
      Q_in = np.sum(Q[Q > 0])
      Q_out = np.sum(Q[Q < 0])
      if vector is not None:
         s = sum(d**2 for d in vector)**0.5        
         n = np.empty((1,1,3), dtype=np.float64)
         n[0,0,:] = [c/s for c in vector]
         abcd = self.params.get_e_type_planes_abcd(4)[:,:,:3]
         r = np.sum(abcd[e4 - self.mesh.e_type_start_index[4]] * n, axis=2)
         Q_v = Q * r[B]
         Q_v_in = np.sum(Q_v[Q_v > 0])
         Q_v_out = np.sum(Q_v[Q_v < 0])         
      self.info.OK()
      if vector is None: return Q_in, Q_out
      else: return Q_in, Q_out, Q_v_in, Q_v_out


   def save_data_to_h5(self, elements, data_refs, file_name):
      self.info.start("Saving element data into hdf5 file '" + file_name + "'")     
      e = elements[np.argsort(self.mesh.e_gmsh_ids[elements])]      
      d = np.empty((len(e), 4 + len(data_refs)), dtype=np.float64)
      d[:,0] = self.mesh.e_gmsh_ids[e]
      d[:,1:4] = self.params.get_e_T()[e]
      for i, r in enumerate(data_refs, 4):
         v = self.data.e_data[r][e]
         if v.ndim == 2: raise self.ToolKitException("Data canot be vector.")
         d[:,i] = v
      with h5.File(file_name, mode="w") as f:
         f.create_dataset("data", data=d)
      self.info.OK()


   def interpolate_values_of_e_data(self, points, e_data_refs, nbs_by_nodes_num, IDW_exp):
      self.info.start("Interpolating Points data")   
      e_data = [self.data.e_data[r] for r in e_data_refs]
      p_data =[np.full(len(points) if (a:=self.data.e_data[r]).ndim == 1 else (len(points), a.shape[1]), np.nan, dtype=np.float64) for r in e_data_refs]      
      p_elems = self.bands.get_points_elements(points)
      A = np.where(p_elems != self.bands.NO_E)[0]
      points_A, p_elems_A = points[A], p_elems[A]
      if len(nbs_by_nodes_num) > 0:
         e_T = self.params.get_e_T()
         c_e_nbs = self.params.get_e_neighbours(nbs_by_nodes_num, np.unique(p_elems_A))         
         for e_nbs in c_e_nbs.values():
            B = np.isin(p_elems_A, e_nbs[:,0])            
            e2i = {e:i for i, e in enumerate(e_nbs[:,0])}            
            p_AB_Eindices = np.fromiter((e2i[e] for e in p_elems_A[B]), np.int32)
            p_AB_Enbs = e_nbs[p_AB_Eindices]
            p, T = points_A[B,None], e_T[p_AB_Enbs]
            p_w = np.sum((p - T)**2, axis=2) ** (- 0.5 * IDW_exp)
            p_w /= np.sum(p_w, axis=1)[:,None]
            for e_d, p_d in zip(e_data, p_data):     
               if e_d.ndim == 1: 
                  p_d[A[B]] = np.sum(p_w * e_d[p_AB_Enbs], axis=1)                      
               else: 
                  p_d[A[B]] = np.sum(p_w[:,:,None] * e_d[p_AB_Enbs], axis=1)
      else:
         for e_d, p_d in zip(e_data, p_data):
            p_d[A] = e_d[p_elems_A]
      self.info.OK()
      return p_data


   def seve_ASC_file(self, file_name, x_min, y_min, cell_size, value_instead_of_nan, data, num_format):
      if x_min is None: x_min = self.params.x_0
      if y_min is None: y_min = self.params.y_0      
      with open(file_name, "w") as f:
         f.write("ncols        {}\n".format(data.shape[1]))
         f.write("nrows        {}\n".format(data.shape[0]))
         f.write("xllcorner    {}\n".format(x_min))
         f.write("yllcorner    {}\n".format(y_min))
         f.write("cellsize     {}\n".format(cell_size))
         f.write("NODATA_value {}\n".format(value_instead_of_nan))
         data[np.isnan(data)] = value_instead_of_nan
         if num_format == "": row_format = " ".join(data.shape[1] * ["{}"]) + "\n"   
         else: row_format = " ".join(data.shape[1] * ["{:"+num_format+"}"]) + "\n"
         for row in data: f.write(row_format.format(*row))


   def save_msh_nodes_grid_data_file(self, file_name, x_min, y_min, cell_size, e_data_refs, data_rasters, num_format):
      self.info.start("Saving nodes grid data to mesh file: '" + file_name+ "'")
      if x_min is None: x_min = self.params.x_0
      if y_min is None: y_min = self.params.y_0
      y_n, x_n = data_rasters[0].shape[:2]
      x_ = x_min + cell_size * np.arange(x_n, dtype=np.float64)
      y_ = y_min + cell_size * np.arange(y_n, dtype=np.float64)      
      with open(file_name, "w") as f:
         f.write("$MeshFormat\n2.2 0 8\n$EndMeshFormat\n")                
         s = "{} {} {} 0\n"
         f.write("$Nodes\n{:n}\n".format(x_n * y_n))  
         f.writelines(s.format(id, *p) for id, p in enumerate(((x,y) for y in y_ for x in x_)))         
         f.write("$EndNodes\n")
         f.write("$Elements\n{:n}\n".format((x_n - 1)*(y_n - 1)))
         s = "{} 3 2 0 0 {} {} {} {}\n"
         f.writelines(s.format(a:=(x+y*x_n), a, a+1, a+1+x_n, a+x_n) for y in range(y_n - 1) for x in range(x_n - 1))
         f.write("$EndElements\n")
         self.info("Mesh was saved.")
         ids = np.arange(x_n * y_n)                  
         for r, d in zip(e_data_refs, data_rasters):
            if d.ndim == 2:               
               d = d.reshape(-1)
               B = np.logical_not(np.isnan(d))
               f.write('$NodeData\n1\n"{:s}"\n1\n{:s}\n3\n{:n}\n1\n{:n}\n'.format(*r[1:], np.sum(B)))               
               if num_format == "": s = "{} {}\n"
               else: s = "{} {:" + num_format + "}\n"
               f.writelines(s.format(*v) for v in zip(ids[B], d[B]))
            else:
               v_n = d.shape[2]
               d = d.reshape(-1, v_n)
               B = np.logical_not(np.isnan(d).all(axis=1))
               f.write('$NodeData\n1\n"{:s}"\n1\n{:s}\n3\n{:n}\n{:n}\n{:n}\n'.format(*r[1:], v_n, np.sum(B)))
               if num_format == "": s = "{}" + v_n * " {}" + "\n"
               else: s = "{}" + v_n * (" {:" + num_format + "}") + "\n"                  
               f.writelines(s.format(id, *v) for id, v in zip(ids[B], d[B]))
            f.write("$EndNodeData\n")        
            self.info("Quantity: '{:s}', time: {:s}, time step: {:n} was saved.".format(*r[1:]))
      self.info.OK()


   def save_msh_elements_grid_data_file(self, file_name, x_min, y_min, cell_size, e_data_refs, data_rasters, num_format):
      self.info.start("Saving elements grid data to mesh file: '" + file_name+ "'")
      if x_min is None: x_min = self.params.x_0
      if y_min is None: y_min = self.params.y_0
      y_n, x_n = data_rasters[0].shape[0] + 1, data_rasters[0].shape[1] + 1
      x_ = - cell_size/2 + x_min + cell_size * np.arange(x_n, dtype=np.float64)
      y_ = - cell_size/2 + y_min + cell_size * np.arange(y_n, dtype=np.float64)      
      with open(file_name, "w") as f:
         f.write("$MeshFormat\n2.2 0 8\n$EndMeshFormat\n")                
         s = "{} {} {} 0\n"
         f.write("$Nodes\n{:n}\n".format(x_n * y_n))  
         f.writelines(s.format(id, *p) for id, p in enumerate(((x,y) for y in y_ for x in x_)))         
         f.write("$EndNodes\n")
         f.write("$Elements\n{:n}\n".format((x_n - 1)*(y_n - 1)))
         s = "{} 3 2 0 0 {} {} {} {}\n"
         f.writelines(s.format(a:=(x+y*x_n), a, a+1, a+1+x_n, a+x_n) for y in range(y_n - 1) for x in range(x_n - 1))
         f.write("$EndElements\n")
         self.info("Mesh was saved.")
         ids = np.empty((x_n-1)*(y_n-1), dtype=np.int32)
         for i in range(y_n-1):
            ids[i*(x_n-1):(i+1)*(x_n-1)] = np.arange(i*x_n, (i+1)*x_n-1)         
         for r, d in zip(e_data_refs, data_rasters):
            if d.ndim == 2:
               d = d.reshape(-1)
               B = np.logical_not(np.isnan(d))
               f.write('$ElementData\n1\n"{:s}"\n1\n{:s}\n3\n{:n}\n1\n{:n}\n'.format(*r[1:], np.sum(B)))
               if num_format == "": s = "{} {}\n"
               else: s = "{} {:" + num_format + "}\n"
               f.writelines(s.format(*v) for v in zip(ids[B], d[B]))
            else:
               v_n = d.shape[2]
               d = d.reshape(-1, v_n)
               B = np.logical_not(np.isnan(d).all(axis=1))               
               f.write('$ElementData\n1\n"{:s}"\n1\n{:s}\n3\n{:n}\n{:n}\n{:n}\n'.format(*r[1:], v_n, np.sum(B)))
               if num_format == "": s = "{}" + v_n * " {}" + "\n"
               else: s = "{}" + v_n * (" {:" + num_format + "}") + "\n"
               f.writelines(s.format(id, *v) for id, v in zip(ids[B], d[B]))
            f.write("$EndElementData\n")        
            self.info("Quantity: '{:s}', time: {:s}, time step: {:n} was saved.".format(*r[1:]))
      self.info.OK()


   def calc_z_level_grid_data(self, z_level, x_min, x_max, y_min, y_max, cell_size, cell_center, data_refs, nbs_by_nodes_num, IDW_exp):
      self.info.start("Calculating raster values of level with z: {:g} m".format(z_level))
      r, a = self.params, cell_size
      if x_min is None: x_min = r.x_0
      if x_max is None: x_max = r.x_max
      if y_min is None: y_min = r.y_0
      if y_max is None: y_max = r.y_max
      if cell_center:
         x_ = np.arange(x_min + a/2, x_max + a/10, a, dtype=np.float64)
         y_ = np.arange(y_min + a/2, y_max + a/10, a, dtype=np.float64)
      else:
         x_ = np.arange(x_min, x_max + a/10, a, dtype=np.float64)
         y_ = np.arange(y_min, y_max + a/10, a, dtype=np.float64)         
      yx_ = np.empty((len(y_), len(x_), 3), dtype=np.float64)
      yx_[:,:,0] = x_
      yx_[:,:,1] = y_[:,None]
      yx_[:,:,2] = z_level
      values = self.interpolate_values_of_e_data(yx_.reshape(-1, 3), data_refs, nbs_by_nodes_num, IDW_exp)            
      values = [v.reshape(yx_.shape[:2]) if v.ndim == 1 else v.reshape(*yx_.shape[:2], v.shape[1]) for v in values]
      self.info.OK()
      return values


   def calc_depth_grid_data(self, depth, depth_by_DEM, x_min, x_max, y_min, y_max, cell_size, cell_center, data_refs, nbs_by_nodes_num, IDW_exp):
      self.info.start("Calculating raster values of depth: {:g} m".format(depth))
      r, a = self.params, cell_size
      if x_min is None: x_min = r.x_0
      if x_max is None: x_max = r.x_max
      if y_min is None: y_min = r.y_0
      if y_max is None: y_max = r.y_max
      if cell_center:
         x_ = np.arange(x_min + a/2, x_max + a/10, a, dtype=np.float64)
         y_ = np.arange(y_min + a/2, y_max + a/10, a, dtype=np.float64)
      else:
         x_ = np.arange(x_min, x_max + a/10, a, dtype=np.float64)
         y_ = np.arange(y_min, y_max + a/10, a, dtype=np.float64)         
      yx_ = np.empty((len(y_), len(x_), 3), dtype=np.float64)
      yx_[:,:,0] = x_
      yx_[:,:,1] = y_[:,None]
      elev = self.DEM.get_points_elevations if depth_by_DEM else self.bands.get_surf_z
      yx_[:,:,2] = elev(yx_[:,:,:2].reshape(-1,2)).reshape(yx_.shape[:2]) - depth    
      values = self.interpolate_values_of_e_data(yx_.reshape(-1, 3), data_refs, nbs_by_nodes_num, IDW_exp)
      values = [v.reshape(yx_.shape[:2]) if v.ndim == 1 else v.reshape(*yx_.shape[:2], v.shape[1]) for v in values]
      self.info.OK()
      return values


   def save_points_data(self, points, points_names, e_data_refs, nbs_by_nodes_num, IDW_exp, file_name, ref_file_name_col=False, col_sep="\t", dec_comma=False, num_format="g", write_to_protocol=True):
      self.info.start("Calculating values of {:n} points".format(len(points)))      
      values = self.interpolate_values_of_e_data(np.array(points), e_data_refs, nbs_by_nodes_num, IDW_exp)            
      num_f = "{}" if num_format == "" else "{:" + num_format + "}"
      head = ["Source file name"] if ref_file_name_col else []
      head.extend(["Quantity", "Time", "Time step"])
      head.extend(points_names)
      rows = [head]
      for r, p_values in zip(e_data_refs, values):
         cols = [r[0]] if ref_file_name_col else []
         cols += [r[1], r[2].replace(".",",") if dec_comma else r[2], str(r[3])] if ref_file_name_col else [r[1], r[2].replace(".",",") if dec_comma else r[2], str(r[3])]
         cols.extend(num_f.format(v).replace(".",",") if dec_comma else num_f.format(v) for v in p_values)
         rows.append(cols)
      if file_name is not None:
         c_n = len(rows[0])
         c_max = [max(len(r[c]) for r in rows) for c in range(c_n)]
         with open(file_name, "w") as f:
            f.write(col_sep.join([h.ljust(n) for h, n in zip(rows[0], c_max)]) + "\n")
            for row in rows[1:]:
               f.write(col_sep.join([r.rjust(n) for r, n in zip(row, c_max)]) + "\n")
      if write_to_protocol:
         p = []
         for r, row in zip(e_data_refs, rows[1:]):
            p.append(("- Element data", "{{Quantity: {:s}, Time: '{:s}', Step: {:n}, File: {:s}}}".format(*r[1:], r[0])))
            p.append(("  Points values", "{" + ", ".join([n + ": " + v for n, v in zip(points_names,row[4:] if ref_file_name_col else row[3:])]) + "}"))
      else: p = None
      self.info.OK()      
      return p


   def save_points_data_with_depth(self, points_xy, points_depth, points_names, e_data_refs, nbs_by_nodes_num, IDW_exp, file_name, depth_by_DEM=False, ref_file_name_col=False, col_sep="\t", dec_comma=False, num_format="g", write_to_protocol=True):
      self.info.start("Calculating values of {:n} points with depth value instead of z coordinate".format(len(points_xy)))      
      points = np.empty((len(points_xy), 3), np.float64)
      points[:,:2] = np.array(points_xy)                        
      elev = self.DEM.get_points_elevations if depth_by_DEM else self.bands.get_surf_z
      points[:,2] = elev(points[:,:2]) - np.array(points_depth, dtype=np.float64)[:,None]
      values = self.interpolate_values_of_e_data(points, e_data_refs, nbs_by_nodes_num, IDW_exp)            
      num_f = "{}" if num_format == "" else "{:" + num_format + "}"
      head = ["Source file name"] if ref_file_name_col else []
      head.extend(["Quantity", "Time", "Time step"])
      head.extend(points_names)
      rows = [head]
      for r, p_values in zip(e_data_refs, values):
         cols = [r[0]] if ref_file_name_col else []
         cols += [r[1], r[2].replace(".",",") if dec_comma else r[2], str(r[3])] if ref_file_name_col else [r[1], r[2].replace(".",",") if dec_comma else r[2], str(r[3])]
         cols.extend(num_f.format(v).replace(".",",") if dec_comma else num_f.format(v) for v in p_values)
         rows.append(cols)
      if file_name is not None:
         c_n = len(rows[0])
         c_max = [max(len(r[c]) for r in rows) for c in range(c_n)]
         with open(file_name, "w") as f:
            f.write(col_sep.join([h.ljust(n) for h, n in zip(rows[0], c_max)]) + "\n")
            for row in rows[1:]:
               f.write(col_sep.join([r.rjust(n) for r, n in zip(row, c_max)]) + "\n")
      if write_to_protocol:
         p = []
         for r, row in zip(e_data_refs, rows[1:]):
            p.append(("- Element data", "{{Quantity: {:s}, Time: '{:s}', Step: {:n}, File: {:s}}}".format(*r[1:], r[0])))
            p.append(("  Points values", "{" + ", ".join([n + ": " + v for n, v in zip(points_names, row[4:] if ref_file_name_col else row[3:])]) + "}"))
      else: p = None
      self.info.OK()      
      return p


   def select_elements(self, physicals_ids=None, x_min=None, x_max=None, y_min=None, y_max=None, z_min=None, z_max=None, depth_min=None, depth_max=None, depth_by_DEM=False, center_in_range=True):
      self.info.start("Selecting elements")
      m = self.mesh
      e = np.arange(m.get_e_count(), dtype=np.int32) if physicals_ids is None else m.get_elements_of_physicals(physicals_ids)   
      if all((x_min is None, y_min is None, z_min is None, depth_min is None, x_max is None, y_max is None, z_max is None, depth_max is None)):
         elems = e     
      elif center_in_range:
         T = self.params.get_e_T()[e]         
         B = np.ones(len(e), dtype=np.bool)
         if x_min is not None: B &= T[:,0] >= x_min
         if y_min is not None: B &= T[:,1] >= y_min
         if z_min is not None: B &= T[:,2] >= z_min
         if x_max is not None: B &= T[:,0] <= x_max
         if y_max is not None: B &= T[:,1] <= y_max
         if z_max is not None: B &= T[:,2] <= z_max
         if depth_min is not None or depth_max is not None:
            if depth_by_DEM: D = self.DEM.get_points_elevations(T[:,:2]) - T[:,2]
            else: D = self.bands.get_surf_z(T[:,:2]) - T[:,2]
            if depth_min is not None: B &= D >= depth_min
            if depth_max is not None: B &= D <= depth_max
         elems = e[B]
      else: 
         P1, P2 = np.empty((len(e),3), np.float64), np.empty((len(e),3), np.float64)
         for t, c in enumerate(m.E_type_nodes_count):
            A = m.choose_e_type(t, e, bool_indices=True)
            if np.any(A):
               d = m.nodes_xyz[m.e_nodes[e[A],:c]]
               P1[A], P2[A] = np.min(d, axis=1), np.max(d, axis=1)
         B = np.ones(len(e), dtype=np.bool)
         if x_min is not None: B &= P2[:,0] >= x_min
         if x_max is not None: B &= P1[:,0] <= x_max
         if y_min is not None: B &= P2[:,1] >= y_min
         if y_max is not None: B &= P1[:,1] <= y_max
         if z_min is not None: B &= P2[:,2] >= z_min
         if z_max is not None: B &= P1[:,2] <= z_max
         if depth_min is not None or depth_max is not None:
            n = (a:=np.unique(m.e_nodes[e].reshape(-1)))[a!=m.NO_NODE_INDEX]            
            if depth_by_DEM: D = self.DEM.get_points_elevations(m.nodes_xyz[n,:2]) - m.nodes_xyz[n,2]
            else: D = self.bands.get_surf_z(m.nodes_xyz[n,:2] ) - m.nodes_xyz[n,2]
            nodes_D = np.empty(len(m.e_nodes), dtype=np.float64)
            nodes_D[n] = D
            D1, D2 = np.empty(len(e), np.float64), np.empty(len(e), np.float64)
            for t, c in enumerate(m.E_type_nodes_count):               
               A = m.choose_e_type(t, e, bool_indices=True)
               if np.any(A):
                  d = nodes_D[m.e_nodes[e[A],:c]]
                  D1[A], D2[A] = np.min(d, axis=1), np.max(d, axis=1)
            if depth_min is not None: B &= D2 >= depth_min
            if depth_max is not None: B &= D1 <= depth_max
         elems = e[B]
      self.info.OK()    
      return elems


   # ************* statistics ***********************

   class StatRec: pass
               
   def get_e_data_statistics(self, elements=None, weights_type=None, data_refs=None, quantiles=None, hist_bins=None, hist_as_pdf=False):
      self.info.start("Calculating statistics of element data")
      if elements is None: elements = np.arange(0, self.mesh.get_count(), dtype=np.int32)
      e_T = self.params.get_e_T()[elements]     
      w_type = weights_type.upper() if weights_type is not None else None
      if w_type is None: e_w = None
      elif w_type == "VOLUME": e_w = self.params.calc_e_volumes(elements)
      elif w_type == "AREA": e_w = self.params.calc_e_areas(elements)
      else: raise self.ToolKitException("Invalid weight type error")
      if data_refs is None: 
         data_refs = sorted(self.data.e_data, key=lambda r: (r[0], r[1], r[3]))      
      if type(quantiles) is int:
         Q = np.linspace(0, 100, quantiles) 
      elif type(quantiles) in (list, tuple, np.ndarray):
         Q = np.array(quantiles)
      if type(hist_bins) is int:
         v1 = min(np.nanmin(self.data.e_data[r][elements]) for r in data_refs)
         v2 = max(np.nanmax(self.data.e_data[r][elements]) for r in data_refs)
         H_edges = np.linspace(v1, v2, hist_bins)
      elif type(hist_bins) in (list, tuple, np.ndarray):
         H_edges = np.array(hist_bins)
      R = []
      for ref in data_refs:
         data = self.data.e_data[ref][elements]
         if data.ndim == 2: raise self.ToolKitException("Quantity '"+str(ref)+ "' canot be vector.")
         r = self.StatRec()
         r.data_ref = ref
         B = np.isfinite(data)
         data = data[B]
         r.weights_type = w_type
         if w_type is None: d_w, r.weights_sum = None, len(data)
         else: d_w, r.weights_sum = e_w[B], np.sum(e_w[B])
         r.min = np.min(data)
         r.max, r.max_position = data[i:=np.argmax(data)], e_T[i]
         r.mean = np.average(data, weights=d_w)
         if quantiles is None: r.Q = r.Q_values = None
         else: r.Q, r.Q_values = Q, np.quantile(data, Q / 100, weights=d_w, method="inverted_cdf")
         if hist_bins is None: r.hist_edges = r.hist_values = None
         else: r.hist_values, r.hist_edges = np.histogram(data, H_edges, weights=d_w)
         if not hist_as_pdf or hist_bins is None: r.pdf_edges = r.pdf_values = None
         else: r.pdf_edges, r.pdf_values = r.hist_edges, 100 * r.hist_values / r.weights_sum
         R.append(r)
      self.info.OK()
      return R


   def print_statistics_as_protocol_results(self, stats_res, num_format="g"):
      p = []
      value_format = ("{:" + num_format + "}") if num_format != "" else "{}"
      w = lambda k, v: p.append(("  "+k if k[0] != "-" else k, v))
      for r in stats_res:
         w("- data file", r.data_ref[0])
         w("Quantity", r.data_ref[1])
         w("Time", r.data_ref[2])
         w("Time step", r.data_ref[3])
         w("Minimum value", r.min)
         w("Maximum value", r.max)
         w("Maximum position", r.max_position)
         w("Mean", r.mean)
         w("Data count" if r.weights_type is None else "Weights summary ("+r.weights_type+")", r.weights_sum)
         if r.Q is not None:
           w("quantiles [%]", "[" + ", ".join([value_format.format(q) for q in r.Q]) + "]")
           w("quantiles_values", "[" + ", ".join([value_format.format(v) for v in r.Q_values]) + "]")
         if r.hist_edges is not None:
           w("histogram edges", "[" + ", ".join([value_format.format(g) for g in r.hist_edges]) + "]")
           w("histogram values", "[" + ", ".join(["{:n}".format(v) for v in r.hist_values]) + "]")
         if r.pdf_edges is not None:
           w("PDF edges", "[" + ", ".join([value_format.format(g) for g in r.pdf_edges]) + "]")
           w("PDF values [%]", "[" + ", ".join([value_format.format(v) for v in r.pdf_values]) + "]")                           
      return p


   def print_statistics_as_table(self, file_name, stats_res, include_data_file_name=True, data_as_columns=False, col_sep=";", dec_comma=False, num_format="g"):
      pf = vf = lambda x: ("{:" + num_format + "}").format(x) if num_format != "" else str(x)
      # ********* head *****************************
      r = stats_res[0]
      h1 = ["Data File"] if include_data_file_name else []
      h2 = [""] if include_data_file_name else []
      h1 += ["Quantity", "Time", "Step", "Mean", "Minimum", "Maximum", "Max. Pos.", "", "", "Weights summary"]      
      h2 += ["", "", "", "", "", "", "x", "y", "z", "Data count" if r.weights_type is None else r.weights_type]      
      if r.Q is not None:
         h1.append("Quantiles [%]")
         if (n:=len(r.Q)) > 1: h1 += (n-1)*[""]
         h2 += [pf(q) for q in r.Q]
      if r.hist_edges is not None:
         h1 += ["Histogram", ""]
         if (n:=len(r.hist_edges)) > 2: h1 += (n-2)*[""]
         h2 += [vf(h) for h in r.hist_edges]
         if r.pdf_edges is not None:
            h1 += ["PDF [%]", ""]
            if n > 2: h1 += (n-2)*[""]
            h2 += [vf(p) for p in r.pdf_edges]
      # ********* data *******************************
      rows = [h1, h2]
      for r in stats_res:
         row = [r.data_ref[0]] if include_data_file_name else []
         row += [*r.data_ref[1:3], str(r.data_ref[3]), vf(r.mean), vf(r.min), vf(r.max)]
         row += ["{:.1f}".format(p) for p in r.max_position]
         row += [str(int(r.weights_sum)) if r.weights_type is None else vf(r.weights_sum)]
         if r.Q is not None:
            row += [vf(v) for v in r.Q_values]
         if r.hist_edges is not None:
            if r.weights_type is None: row += [str(int(v)) for v in r.hist_values] + [""]
            else: row += [vf(v) for v in r.hist_values] + [""]
            if r.pdf_edges is not None:
               row += [pf(p) for p in r.pdf_values] + [""]
         if dec_comma:
            for i in range(4 if include_data_file_name else 3, len(row)):
               row[i] = row[i].replace(".", ",")
         rows.append(row)
      A = np.array(rows, dtype=object)
      if data_as_columns: A = A.transpose()
      # ********** write to file *********************
      col_max_len = [max(len(s) for s in A[:,c]) for c in range(A.shape[1])]
      lines = []
      if data_as_columns:
         for row in A:
            h = col_sep.join([s.ljust(n) for s, n in zip(row[:2], col_max_len[:2])])
            d = col_sep.join([s.rjust(n) for s, n in zip(row[2:], col_max_len[2:])])
            lines.append(h + col_sep + d)
      else:
         for row in A[:2]:
            lines.append(col_sep.join([s.ljust(n) for s, n in zip(row, col_max_len)]))
         for row in A[2:]:
            lines.append(col_sep.join([s.rjust(n) for s, n in zip(row, col_max_len)]))
      with open(file_name, "w") as f:
         for s in lines: f.write(s + "\n")
      return lines
