import tkinter as tk
from tkinter import filedialog
import subprocess as sp
import os, psutil

from diskmod_tools import DiskmodTools

from workflow_GUI import BaseApplication
from workflow_GUI import theme
from workflow_GUI.messagebox import msg_info, msg_error, msg_yesno
from workflow_GUI.dlg_wait_for_actions import get_wait_for_actions_ids_dialog

from dialogs.open_mesh import OpenMeshDialog
from dialogs.entity_select import EntitySelectionDialog
from dialogs.entity_cross_section import EntityCrossSectionDialog
from dialogs.reference_point import ReferencePointDialog
from dialogs.show_model_range import ShowModelRangeDialog
from dialogs.show_entities import ShowEntitiesDialog
from dialogs.save_mesh import SaveMeshDialog
from dialogs.top_boundary_z_change import TopBoundaryZChangeDialog
from dialogs.level_above_z_change import LevelAboveZChangeDialog
from dialogs.merge_entities import MergeEntitiesDialog
from dialogs.change_entity_id_name import ChangeEntityNameIdDialog
from dialogs.open_output_data import OpenOutputDataDialog
from dialogs.show_data import ShowDataDialog
from dialogs.save_data import SaveDataDialog
from dialogs.set_boundary_data import SetBoundaryDataDialog
from dialogs.flux_between_entities import FluxBetweenEntitiesDialog
from dialogs.data_statistics import DataStatisticsDialog
from dialogs.data_depth_z_raster import DataDepthZlevelRasterDialog
from dialogs.data_points import DataPointsDialog
from dialogs.data_vectors_to_scalars import DataVectorstoScalarsDialog
from dialogs.data_by_formula import DataByFormulaDialog


class DiskmodGUIApplication(BaseApplication):

   def __init__(self, option_file_name):
      super().__init__(option_file_name)
      self.set_app_name("DISKMODtools")
      try: self.iconbitmap(self.extent_file_name("DM-32.ico"))
      except: print("'DM-32.ico' opening error")
                        
      # *********** bar ******************
      self.init_toolbar()
      self.init_statusbar()
      
      # *************** menu ***************
      self.create_menu_model()
      self.create_menu_mesh_edit()
      self.create_menu_data_manipulation()
      self.create_menu_data_evaluation()
      self.create_menu_show()
      self.create_menu_help()


   def do_after_protocol_delete(self):
      self.diskmod = DiskmodTools(self.info)

   def do_after_protocol_perform(self):
      # self.on_set_reference_point()
      pass
       
      
   def do_refresh_status(self):      
      p, m = psutil.Process(), psutil.virtual_memory()     
      s = " Memory - used by app: {:_} MB, available: {:_} MB ".format(p.memory_info().rss // 2**20, m.available // 2**20)   
      self.status_text["text"] = s.replace("_", " ")


   # ************* statusbar **************************

   def init_toolbar(self):
      t = self.toolbar
      b = theme.ToolBar_Button(t, text="Mesh", command=self.on_open_mesh_file)
      b.grid(row=0, column=1)
      b = theme.ToolBar_Button(t, text="DEM", command=self.on_open_DEM_file)
      b.grid(row=0, column=2)
      b = theme.ToolBar_Button(t, text="Top Boundary", command=self.on_top_boundary_entity_selection)
      b.grid(row=0, column=3)
      b = theme.ToolBar_Button(t, text="Domain Range", command=self.on_show_model_range)
      b.grid(row=0, column=4)
      b = theme.ToolBar_Button(t, text="Entities Summary", command=self.on_show_entities)
      b.grid(row=0, column=5)
      b = theme.ToolBar_Button(t, text="Show by GMSH", command=self.on_show_by_gmsh)
      b.grid(row=0, column=6)
      b = theme.ToolBar_Button(t, text="Show by default", command=self.on_show_by_default)
      b.grid(row=0, column=7)
      

      
      
   def init_statusbar(self):
      self.status_text = theme.Label(self.statusbar)
      self.status_text.grid(row=0, column=0, sticky="W")


   # ***************** menu **********************


   def create_menu_model(self):
      self.mainmenu.add_cascade(label="Model Basics", menu=(m:=theme.Menu(self.mainmenu, tearoff=0)))
      self.add_actions_menu_items(m, [
        ("Load Mesh File", "open_mesh_file", None),
        ("Load DEM File", "open_DEM_file", None),
        ("Top Boundary Entity Selection", "top_boundary_entity_selection", None),
        ("Set Entity Cross Section", "set_entity_cross_section", None),
        ("Set Reference XY Position", "set_reference_point", None)
      ])


   def create_menu_mesh_edit(self):
      self.mainmenu.add_cascade(label="Mesh Edit", menu=(m:=theme.Menu(self.mainmenu, tearoff=0)))
      self.add_actions_menu_items(m, [
        ("Model Domain Range Summary", "show_model_range", None),
        ("Model Entities Summary", "show_entities", None),
        ("-", None, None),
        ("Change of Z-coordinates on Top Boundary Only", "change_top_boundary_z_coord", None),
        ("Change of Z-coordinates from the Specified Level", "change_level_above_z_coord", None),
        ("-", None, None),
        ("Remove Entities", "remove_entities", None),
        ("Merge Entities", "merge_entities", None),
        ("Change Entities Name/Number", "change_entity_name_id", None),
        ("-", None, None),
        ("Save Modified Mesh", "save_mesh", None)  ])


   def create_menu_data_manipulation(self):
      self.mainmenu.add_cascade(label="Data Management", menu=(m:=theme.Menu(self.mainmenu, tearoff=0)))
      self.add_actions_menu_items(m, [
        ("Load Element Data of Identical Mesh", "open_element_data_of_same_mesh", None),
        ("Load Element Data of Any Mesh", "open_element_data_of_any_mesh", None),
        ("Assign Data to Boudary Elements", "set_boundary_data", None),
        ("Loaded Element Data Summary", "show_element_data", None),
        ("-", None, None),
        ("Select and Save Element Data", "save_partial_data", None),
        ("Select and Save Element Data With Complete Mesh", "save_partial_data_with_complete_mesh", None),
        ("-", None, None),
        ("Load Water Flow Raw", "open_water_flow_raw", None)
      ])


   def create_menu_data_evaluation(self):
      self.mainmenu.add_cascade(label="Data Processing", menu=(m:=theme.Menu(self.mainmenu, tearoff=0)))
      self.add_actions_menu_items(m, [
        ("Analyse Flux Between Entities", "calc_flux_between_entities", None),
        ("-", None, None),
        ("Analyse Element Data Statistics", "calc_data_statistics", None),
        ("Vector Element Data to Scalar Data", "create_scalar_data_from_vectors", None),
        ("Create New Element Data by Formula", "calc_data_by_formula", None),
        ("-", None, None),
        ("Save Z-level Raster", "get_z_level_raster", None),
        ("Save Depth Raster", "get_depth_raster", None),
        ("Save Point Data", "get_data_points", None),
        ("Save Point Data of Depth", "get_data_points_with_depth", None)
      ])


   def create_menu_show(self):      
      sm = [("Show Selected Filepath by GMSH", "show_by_gmsh", None),
            ("Show Selected Filepath by Notepad (Windows)", "show_by_notepad", None) ]
      A = self.options["applications"]
      if (s:="Microsoft Excel") in A:
         if A[s]["use"]: 
            sm.append(("Show Selected Filepath by Microsoft Excel", "show_by_librecalc", None))
      if (s:="Libre Calc") in A:
         if A[s]["use"]: 
            sm.append(("Show Selected Filepath by Libre Calc", "show_by_librecalc", None))
      sm.append(("-", None, None))
      sm.append(("Show File by Default Application", "show_by_default", None))
      self.mainmenu.add_cascade(label="Show", menu=(m:=theme.Menu(self.mainmenu, tearoff=0)))
      self.add_actions_menu_items(m, sm)
     

   def create_menu_help(self):
      self.mainmenu.add_cascade(label="Help", menu=(m:=theme.Menu(self.mainmenu, tearoff=0)))
      self.add_actions_menu_items(m, [
        ("Documentation", "help_documentation", None),
        ("About Program", "help_about", None) ])


   # ********** help methods ******************

   def is_prepared(self, items):
      m, dm = [], self.diskmod
      if "mesh" in items and not dm.is_mesh_loaded():
         m.append("No mesh is loaded")
      if "DEM" in items and not dm.is_DEM_loaded():
         m.append("Mo DEM is loaded.")
      if "top_boundary" in items and not dm.is_top_boudary_setting():
         m.append("No top boudary is set.")
      if "element_data" in items and not dm.is_element_data():
         m.append("No element data is loaded.")
      if "raw" in items and not dm.is_water_flow_raw_loaded():
         m.append("No water flow raw file is loaded.")         
      if len(m) == 0: return True
      else:
         msg_info(self, "\n".join(m))
         return False



   # ***************************************************
   # *********** Model Basics **************************

   # *********** open mesh *********************

   def on_open_mesh_file(self):
      action_name = self.get_action_name_by_myname()
      d = OpenMeshDialog(self, action_name)
      if d.execute_modal():
         self.ex_open_mesh_file((self.gen_action_id(), action_name, d.out_wfa_ids),
            d.mesh_file.get(), 
            p if (p:=d.physical_names_file.get()) != "" else None)

   def re_open_mesh_file(self, id_name_wait, rec):
      p = rec[self.pk_inputs]
      self.ex_open_mesh_file(id_name_wait,
         p["Mesh file"], p["Physical names file"] )

   def ex_open_mesh_file(self, id_name_wait, mesh_file, physical_names_file):
      f = self.create_action_frame(*id_name_wait,
         self.diskmod.load_mesh, mesh_file, physical_names_file)
      f.start_with_default_protocoling(
         [("Mesh file", mesh_file), ("Physical names file", physical_names_file)],
         None ) 
           
      
   # ********** open DEM ***********************
   
   def on_open_DEM_file(self):
      action_name = self.get_action_name_by_myname()
      file_name = filedialog.askopenfilename(
         title=action_name,
         filetypes=[("Elevations grid file (*.asc)", ".asc")],
         defaultextension=".asc",
         initialdir=".")
      if file_name != "":
         self.ex_open_DEM_file((self.gen_action_id(), action_name, None),
            file_name)

   def re_open_DEM_file(self, id_name_wait, rec):
      p = rec[self.pk_inputs]
      self.ex_open_DEM_file(id_name_wait, p["DEM file"] )

   def ex_open_DEM_file(self, id_name_wait, file_name):
      f = self.create_action_frame(*id_name_wait,
         self.diskmod.load_DEM, file_name)
      f.start_with_default_protocoling(
         [("DEM file", file_name)], None )


   # ********** physical entity selection *************

   def on_top_boundary_entity_selection(self): 
      if not self.is_prepared("mesh"): return
      action_name = self.get_action_name_by_myname()
      if (thread:=self.perform_isolated_action(action_name, self.diskmod.get_top_boundary_physicals_selection)) is None: return
      d = EntitySelectionDialog(self, action_name, allow_no_selection=True)
      d.set_inputs(thread.result[0], thread.result[1])
      if d.execute_modal():
         f = self.create_action_frame(self.gen_action_id(), thread.action_name, d.out_wfa_ids,
            self.diskmod.set_top_boundary_physicals_selection, d.out_selected_ids)
         f.start_with_option_protocoling(
            [("Selected entity ID", d.out_selected_ids)] )
            
   def re_top_boundary_entity_selection(self, id_name_wait, rec):       
      ids = rec[self.pk_options]["Selected entity ID"]
      f = self.create_action_frame(*id_name_wait,
         self.diskmod.set_top_boundary_physicals_selection, ids)
      f.start_with_option_protocoling(
         [("Selected entity ID", ids)] )


   # ******* entity cross section ****************

   def on_set_entity_cross_section(self):
      if not self.is_prepared("mesh"): return
      action_name = self.get_action_name_by_myname()
      if (thread:=self.perform_isolated_action("Get physicals names params", self.diskmod.get_physicals_names_cross_sections)) is None: return  
      d = EntityCrossSectionDialog(self, action_name)
      d.set_inputs(*thread.result)
      if d.execute_modal():
         self.ex_set_entity_cross_section((self.gen_action_id(), action_name, d.out_wfa_ids),
            d.out_entities, d.out_default)
      
   def re_set_entity_cross_section(self, id_name_wait, rec):
      p = rec[self.pk_options]
      d = p["Default cross section"]
      e = {t["ID"]:t["Cross section"] for t in p["Selected entities"]}
      self.ex_set_entity_cross_section(id_name_wait, e, d)
 
   def ex_set_entity_cross_section(self, id_name_wait, entity_values, default_value):
      f = self.create_action_frame(*id_name_wait,
         self.diskmod.set_physicals_names_cross_section, entity_values, default_value)      
      if len(entity_values) > 0:
         f.start_with_option_protocoling(
            [("Default cross section", "{:g}".format(default_value)),
             ("Selected entities", ""),
             *(("  - ", "{ID: "+str(id)+", Cross section: {:g}".format(c)+"}") for id, c in sorted(entity_values.items())) ] )      
      else:
         f.start_with_option_protocoling(
            [("Default cross section", "{:g}".format(default_value))] )
   

   # *************** reference position ************
   
   def on_set_reference_point(self):
      if not self.is_prepared("mesh"): return
      action_name = self.get_action_name_by_myname()
      if (thread:=self.perform_isolated_action(action_name, self.diskmod.get_reference_point)) is None: return
      d = ReferencePointDialog(self, action_name)
      d.set_inputs(*thread.result)
      if d.execute_modal():
         self.ex_set_reference_point((self.gen_action_id(), action_name, d.out_wfa_ids), d.out_x_ref, d.out_y_ref)

   def re_set_reference_point(self, id_name_wait, rec):
      p = rec[self.pk_options]
      self.ex_set_reference_point(id_name_wait, p["X"], p["Y"])

   def ex_set_reference_point(self, id_name_wait, x, y):
      f = self.create_action_frame(*id_name_wait, self.diskmod.set_reference_point, x, y)
      f.start_with_option_protocoling(
         options=[("X", x), ("Y", y)] )
       

   # *****************************j***********************
   # **************** Model edit ************************

   # ********* show model range *************

   def on_show_model_range(self):
      if not self.is_prepared("mesh"): return
      action_name = self.get_action_name_by_myname()
      if (thread:=self.perform_isolated_action(action_name, self.diskmod.report_model_range)) is None: return
      r = thread.result
      d = ShowModelRangeDialog(self, action_name)
      d.set_inputs(r[0], r[2])
      if d.execute_modal():
         if d.out_write_to_protocol:
            thread.action_id = self.gen_action_id()
            p = [(k, self.float2str(v)) for k, v in (r[1] if r[3] is None else r[1] + r[3])]
            self.protocol_perform_action(thread, results=p)
            

   # ******** show entities *************************

   def on_show_entities(self):
      if not self.is_prepared("mesh"): return
      action_name = self.get_action_name_by_myname()
      if (thread:=self.perform_isolated_action(action_name, self.diskmod.get_physicals_names)) is None: return
      d = ShowEntitiesDialog(self, action_name)
      d.set_inputs(thread.result, [])
      d.execute_modal()


   # ******** change top boudary z coordinates ******

   def on_change_top_boundary_z_coord(self):
      if not self.is_prepared(["mesh","top_boundary","DEM"]): return
      action_name = self.get_action_name_by_myname()
      d = TopBoundaryZChangeDialog(self, action_name)
      if d.execute_modal():
         self.ex_change_top_boundary_z_coord((self.gen_action_id(), action_name, d.out_wfa_ids), d.out_file_name)

   def re_change_top_boundary_z_coord(self, id_name_wait, rec):
      p = rec.get(self.pk_inputs)
      file_name = p["Result file name"] if p is not None else None
      self.ex_change_top_boundary_z_coord(id_name_wait, file_name)

   def ex_change_top_boundary_z_coord(self, id_name_wait, diff_file_name):
      f = self.create_action_frame(*id_name_wait,
         self.diskmod.change_top_nodes_z_by_DEM, diff_file_name)
      f.start_with_default_protocoling(
         None if diff_file_name is None else [("Result file name", diff_file_name)],
         None ) 


   # ******** change level above z coordinates ******

   def on_change_level_above_z_coord(self):
      if not self.is_prepared(["mesh","top_boundary","DEM"]): return
      action_name = self.get_action_name_by_myname()
      d = LevelAboveZChangeDialog(self, action_name)
      if d.execute_modal():
         self.ex_change_level_above_z_coord((self.gen_action_id(), action_name, d.out_wfa_ids), d.out_z_level, d.out_file_name)

   def re_change_level_above_z_coord(self, id_name_wait, rec):
      p = rec[self.pk_inputs]
      z_level = p["Z Level"]
      file_name = p[k] if (k:="Result file name") in p else None
      self.ex_change_level_above_z_coord(id_name_wait, z_level, file_name)

   def ex_change_level_above_z_coord(self, id_name_wait, z_level, diff_file_name):
      f = self.create_action_frame(*id_name_wait,
         self.diskmod.change_nodes_z_from_level_by_DEM, z_level, diff_file_name)
      inputs = [("Z Level", z_level)]
      if diff_file_name is not None: 
         inputs.append(("Result file name", diff_file_name))
      f.start_with_default_protocoling(
         inputs, 
         None ) 


   # *************** remove entities ****************

   def on_remove_entities(self):
      if not self.is_prepared("mesh"): return
      action_name = self.get_action_name_by_myname()
      if (thread:=self.perform_isolated_action(action_name, self.diskmod.get_physicals_names)) is None: return
      d = EntitySelectionDialog(self, action_name)
      d.set_inputs(thread.result, [])
      if d.execute_modal():
         self.ex_remove_entities((self.gen_action_id(), action_name, d.out_wfa_ids), d.out_selected_ids)

   def re_remove_entities(self, id_name_wait, rec):
      p = rec[self.pk_inputs]
      self.ex_remove_entities(id_name_wait, p["Entities IDs"])

   def ex_remove_entities(self, id_name_wait, physicals_ids):
      f = self.create_action_frame(*id_name_wait, self.diskmod.remove_physicals, physicals_ids)
      f.start_with_default_protocoling(
         [("Entities IDs", physicals_ids)], 
         None ) 


   # *************** merge entities ****************

   def on_merge_entities(self):
      if not self.is_prepared("mesh"): return
      action_name = self.get_action_name_by_myname()
      if (thread:=self.perform_isolated_action(action_name, self.diskmod.get_physicals_names)) is None: return
      d = MergeEntitiesDialog(self, action_name)
      d.set_inputs(thread.result, [])
      if d.execute_modal():
         self.ex_merge_entities((self.gen_action_id(), action_name, d.out_wfa_ids), 
              d.out_name, d.out_id, d.out_selected_ids)

   def re_merge_entities(self, id_name_wait, rec):
      p = rec[self.pk_inputs]
      self.ex_merge_entities(id_name_wait, p["New entity name"], p["New entity ID"], p["Entities IDs"])

   def ex_merge_entities(self, id_name_wait, new_name, new_id, physicals_ids):
      f = self.create_action_frame(*id_name_wait, self.diskmod.merge_physicals, new_name, new_id, physicals_ids)
      f.start_with_default_protocoling(
         [("Entities IDs", physicals_ids), 
          ("New entity ID", new_id), ("New entity name", new_name)],
          None )
       
         
   # ************ change entity id, name ******************

   def on_change_entity_name_id(self):
      if not self.is_prepared("mesh"): return
      action_name = self.get_action_name_by_myname()
      d = ChangeEntityNameIdDialog(self, action_name)
      d.execute_modal()
      

   # *************** save mesh *********************

   def on_save_mesh(self):
      if not self.is_prepared("mesh"): return
      action_name = self.get_action_name_by_myname()
      d = SaveMeshDialog(self, action_name)
      if d.execute_modal():
         self.ex_save_mesh_file((self.gen_action_id(), action_name, d.out_wfa_ids),
            d.out_mesh_file )
      
   def re_save_mesh(self, id_name_wait, rec):
      p = rec[self.pk_inputs]
      self.ex_save_mesh_file(id_name_wait, p["Mesh file"])

   def ex_save_mesh_file(self, id_name_wait, mesh_file):
      f = self.create_action_frame(*id_name_wait,
         self.diskmod.save_mesh, mesh_file)
      f.start_with_default_protocoling(
         [("Mesh file", mesh_file)],
         None ) 


   # *****************************************************
   # *****************element data ***********************

   def protocol_data_refs(self, title, data_refs, indent=""):
      if data_refs is None: s = [(indent + title, "all")]
      elif len(data_refs) == 0: s = [(indent + title, "no data")]
      elif len(data_refs) == 1: s = [(indent + title, "{{Quantity: {:s}, Time: '{:s}', Step: {:n}, File: {:s}}}".format(*(data_refs[0][1:]), data_refs[0][0]))]
      else: s = [(indent + title, "")] + [(indent + "  - ", "{{Quantity: {:s}, Time: '{:s}', Step: {:n}, File: {:s}}}".format(*a[1:], a[0])) for a in data_refs]
      return s
   
   def get_protocol_e_data_refs(self, data_refs, indent=""):
      return self.protocol_data_refs("Element data", data_refs, indent=indent)

   def data_refs_from_protocol(self, dict_with_data, data_key_name):
      if (p := dict_with_data[data_key_name]) == "all": r = None      
      elif p == "no data": r = []
      elif type(p) is dict: r = [(p["File"], p["Quantity"], p["Time"], p["Step"])]
      else: r = [(d["File"], d["Quantity"], d["Time"], d["Step"]) for d in p]
      return r
      
   def e_data_refs_from_protocol(self, dict_with_data):
      return self.data_refs_from_protocol(dict_with_data, "Element data")


   # *********** open element data from same mesh *********

   def get_protocol_loaded_e_data_refs(self, data_refs):
      return self.protocol_data_refs("Loaded element data", data_refs)

   def on_open_element_data_of_same_mesh(self):
      if not self.is_prepared("mesh"): return
      action_name = self.get_action_name_by_myname()
      d = OpenOutputDataDialog (self, action_name)
      if d.execute_modal():
         self.ex_open_element_data_of_same_mesh((self.gen_action_id(), action_name, d.out_wfa_ids),
            d.out_msh_file, d.out_names, d.out_times, d.out_steps)

   def re_open_element_data_of_same_mesh(self, id_name_wait, rec):
      p = rec[self.pk_inputs]
      f, n, t, s = p["Mesh file"], p["Names"], p["Times"], p["Steps"]
      self.ex_open_element_data_of_same_mesh(id_name_wait, f, n if n!="all" else None, t if t!="all" else None, s if s!="all" else None)

   def ex_open_element_data_of_same_mesh(self, id_name_wait, mesh_file, names, times, times_steps):
      f = self.create_action_frame(*id_name_wait,
         self.diskmod.load_element_data_from_same_mesh , mesh_file, names=names, times=times, times_indices=times_steps)
      f.start_with_default_protocoling(
         [("Mesh file", mesh_file), ("Names", "all" if names is None else names), ("Times", "all" if times is None else times), ("Steps","all" if times_steps is None else times_steps)],       
         self.get_protocol_loaded_e_data_refs )
         

   # *********** open element data from any mesh *********

   def on_open_element_data_of_any_mesh(self):
      if not self.is_prepared("mesh"): return
      action_name = self.get_action_name_by_myname()
      d = OpenOutputDataDialog (self, action_name)
      if d.execute_modal():
         self.ex_open_element_data_of_any_mesh((self.gen_action_id(), action_name, d.out_wfa_ids),
            d.out_msh_file, d.out_names, d.out_times, d.out_steps)

   def re_open_element_data_of_any_mesh(self, id_name_wait, rec):
      p = rec[self.pk_inputs]
      f, n, t, s = p["Mesh file"], p["Names"], p["Times"], p["Steps"]
      self.ex_open_element_data_of_any_mesh(id_name_wait, f, n if n!="all" else None, t if t!="all" else None, s if s!="all" else None)

   def ex_open_element_data_of_any_mesh(self, id_name_wait, mesh_file, names, times, times_steps):
      f = self.create_action_frame(*id_name_wait,
         self.diskmod.load_element_data_from_any_mesh , mesh_file, names=names, times=times, times_indices=times_steps)
      f.start_with_default_protocoling(
         [("Mesh file", mesh_file), ("Names","all" if names is None else names), ("Times","all" if times is None else times), ("Steps","all" if times_steps is None else times_steps)],
         self.get_protocol_loaded_e_data_refs )

   # *********** show element data ***************
   
   def on_show_element_data(self):
      if not self.is_prepared(["mesh","element_data"]): return
      action_name = self.get_action_name_by_myname()
      if (thread:=self.perform_isolated_action(action_name, self.diskmod.get_element_data_info)) is None: return
      d = ShowDataDialog(self, action_name)
      d.set_inputs(thread.result)      
      d.execute_modal()
               
   def re_show_element_data(self, id_name_wait, rec):
      p = rec[self.pk_inputs]
      data_ref = self.e_data_refs_from_protocol(p)
      f = self.create_action_frame(*id_name_wait,
         self.diskmod.get_element_data_info, data_ref)
      f.start_with_default_protocoling(
         self.get_protocol_e_data_refs(data_ref),
         lambda r: r[data_ref[0]].protocol )
      
      

   # *********** set boundary data ******************
   
   def on_set_boundary_data(self):
      if not self.is_prepared(["mesh","element_data"]): return
      action_name = self.get_action_name_by_myname()
      if (thread:=self.perform_isolated_action(action_name, self.diskmod.get_element_data_references_physicals_names)) is None: return
      d = SetBoundaryDataDialog(self, action_name)
      d.set_inputs(thread.result[1], thread.result[0])      
      if d.execute_modal():
         self.ex_set_boundary_data((self.gen_action_id(), action_name, d.out_wfa_ids),
            d.out_sel_physicals_ids ,d.out_sel_data_refs)

   def re_set_boundary_data(self, id_name_wait, rec):
      p = rec[self.pk_inputs]
      e, d = p["Entities IDs"], self.e_data_refs_from_protocol(p)
      self.ex_set_boundary_data(id_name_wait, e, d)

   def ex_set_boundary_data(self, id_name_wait, physicals_ids, data_refs):
      f = self.create_action_frame(*id_name_wait,
         self.diskmod.set_e_data_boundary_e2_physicals_as_e4_neighbours, data_refs, physicals_ids)
      f.start_with_default_protocoling(
         [("Entities IDs", physicals_ids)] + self.get_protocol_e_data_refs(data_refs),
          None )


   # ****** partition params to protocol ******************

   def create_protocol_partition_params(self, physicals_ids, ranges):            
      R = list(ranges.values()).count(None) < 8
      if physicals_ids is None and not R: return []
      p = [("Partition description", "")]
      if physicals_ids is not None: p.append(("  Entities IDs", physicals_ids))
      def w(s, t):
         if ranges.get(s) is not None: p.append(("  "+t, ranges[s]))
      w("x_min", "Minimum x"); w("x_max", "Maximum x")
      w("y_min", "Minimum y"); w("y_max", "Maximum y")
      if any(ranges[s] is not None for s in ("x_min", "x_max", "y_min", "y_max")):
         p.append(("  XY position style", "offset to reference position" if ranges["xy_by_ref"] else "absolute position"))
      w("z_min", "Minimum z"); w("z_max", "Maximum z")
      w("depth_min", "Minimum depth"); w("depth_max", "Maximum depth")     
      if R: p.append(("  Element center included", ranges["center_in_range"]))
      if ranges.get("depth_min") is not None or ranges.get("depth_max") is not None: p.append(("  Depth ralative to", "DEM" if ranges["depth_by_DEM"] else "top boundary entities"))
      return p

   def get_protocol_partition_params(self, inputs_rec):
      if (r:=inputs_rec.get("Partition description")) is None: return None, {}
      physicals_ids = r.get("Entities IDs")
      p = {}
      def w(t, s): 
         if r.get(t) is not None: p[s] = r.get(t)
      w("Minimum x", "x_min"); w("Maximum x", "x_max")
      w("Minimum y", "y_min"); w("Maximum y", "y_max")
      if len(p) > 0:
         p["xy_by_ref"] = r["XY position style"] != "absolute position"
      w("Minimum z", "z_min"); w("Maximum z", "z_max")
      w("Minimum depth", "depth_min"); w("Maximum depth", "depth_max")     
      if len(p) > 0:
         p["center_in_range"] = r["Element center included"]
         if "depth_min" in p or "depth_max" in p:
            p["depth_by_DEM"] = r["Depth relative to"] == "DEM"
      return physicals_ids, p


   # *********** save partial data **************************
   
   def on_save_partial_data(self):
      if not self.is_prepared(["mesh","element_data"]): return
      action_name = self.get_action_name_by_myname()
      if (thread:=self.perform_isolated_action(action_name, self.diskmod.get_element_data_references_physicals_names)) is None: return
      d = SaveDataDialog(self, action_name, self.diskmod.is_DEM_loaded())
      d.set_inputs(thread.result[1], thread.result[0])      
      if d.execute_modal():
         self.ex_save_partial_data((self.gen_action_id(), action_name, d.out_wfa_ids),
            d.out_file_name, d.out_sel_data_refs, d.out_sel_physicals_ids, d.out_range)

   def re_save_partial_data(self, id_name_wait, rec):
      p = rec[self.pk_inputs]      
      f, d = p["File"], self.e_data_refs_from_protocol(p)
      e, r = self.get_protocol_partition_params(p)
      self.ex_save_partial_data(id_name_wait, f, d, e, r)
   
   def ex_save_partial_data(self, id_name_wait, file_name, data_refs, physicals_ids, ranges):
      f = self.create_action_frame(*id_name_wait,
         self.diskmod.save_partial_element_data,
         file_name, data_refs=data_refs, physicals_ids=physicals_ids, **ranges)
      p = [("File", file_name)] 
      p.extend(self.get_protocol_e_data_refs(data_refs))
      p.extend(self.create_protocol_partition_params(physicals_ids, ranges))
      f.start_with_default_protocoling(
         p,
         None )
      

   # ********* save partial data with complete mesh **************************
   
   def on_save_partial_data_with_complete_mesh(self):
      if not self.is_prepared(["mesh","element_data"]): return
      action_name = self.get_action_name_by_myname()
      if (thread:=self.perform_isolated_action(action_name, self.diskmod.get_element_data_references_physicals_names)) is None: return
      d = SaveDataDialog(self, action_name, self.diskmod.is_DEM_loaded())
      d.set_inputs(thread.result[1], thread.result[0])      
      if d.execute_modal():
         self.ex_save_partial_data_with_complete_mesh((self.gen_action_id(), action_name, d.out_wfa_ids),
            d.out_file_name, d.out_sel_data_refs, d.out_sel_physicals_ids, d.out_range)

   def re_save_partial_data_with_complete_mesh(self, id_name_wait, rec):
      p = rec[self.pk_inputs]    
      f, d = p["File"], self.e_data_refs_from_protocol(p)      
      e, r = self.get_protocol_partition_params(p)
      self.ex_save_partial_data_with_complete_mesh(id_name_wait, f, d, e, r)
   
   def ex_save_partial_data_with_complete_mesh(self, id_name_wait, file_name, data_refs, physicals_ids, ranges):
      f = self.create_action_frame(*id_name_wait,
         self.diskmod.save_partial_element_data_with_complete_mesh,
         file_name, data_refs=data_refs, physicals_ids=physicals_ids, **ranges)
      p = [("File", file_name)] 
      p.extend(self.get_protocol_e_data_refs(data_refs))
      p.extend(self.create_protocol_partition_params(physicals_ids, ranges))      
      f.start_with_default_protocoling(
         p,
         None )

      
   # *********** open flow raw ************************

   def on_open_water_flow_raw(self):
      if not self.is_prepared("mesh"): return
      action_name = self.get_action_name_by_myname()
      file_name = filedialog.askopenfilename(
         title=action_name,
         filetypes=[("Flow123d water flow raw (*.txt)", ".txt")],
         defaultextension=".txt",
         initialdir=".")
      if file_name != "":
         if msg_yesno(self, "Do you no select actions to waiting?", title="Waiting to actions"):
            state, wait_ids = get_wait_for_actions_ids_dialog(self)
            if not state: return
         else: wait_ids = None
         self.ex_open_water_flow_raw((self.gen_action_id(), action_name, wait_ids),
            file_name)

   def re_open_water_flow_raw(self, id_name_wait, rec):
      p = rec[self.pk_inputs]
      self.ex_open_water_flow_raw(id_name_wait, p["Flow raw file"])

   def ex_open_water_flow_raw(self, id_name_wait, file_name):
      f = self.create_action_frame(*id_name_wait,
         self.diskmod.load_water_flow_raw, file_name)
      f.start_with_default_protocoling(
         [("Flow raw file", file_name)],
         lambda r: [("Loaded times", list(r))])


   # ********** calc e4 flux between entities *************

   def on_calc_flux_between_entities(self):
      if not self.is_prepared(["mesh", "raw"]): return
      action_name = self.get_action_name_by_myname()
      if (thread:=self.perform_isolated_action(action_name, self.diskmod.get_physicals_names)) is None: return
      d = FluxBetweenEntitiesDialog(self, action_name)      
      d.set_inputs(thread.result)
      if d.execute_modal():
          self.ex_calc_flux_between_entities((self.gen_action_id(), action_name, d.out_wfa_ids),
            d.out_main_entities_ids, d.out_nbs_entities_ids, d.out_T, d.out_vector)

   def re_calc_flux_between_entities(self, id_name_wait, rec):
      p = rec[self.pk_inputs]
      self.ex_calc_flux_between_entities(id_name_wait, p["Main entities IDs"], p["Neighbours entities IDs"], p["Time"], p["Vector"])

   def ex_calc_flux_between_entities(self, id_name_wait, main_ids, nbs_ids, T, vector):
      f = self.create_action_frame(*id_name_wait,
         self.diskmod.calc_e4_flux_over_e2e4_neighbours, main_ids, nbs_ids, T, vector)      
      f.start_with_default_protocoling(
         [("Main entities IDs", main_ids),
          ("Neighbours entities IDs", nbs_ids),
          ("Time", T),
          ("Vector", vector if vector is not None else "") ],
         lambda r : [("Q_in", r[0]), ("Q_out", r[1])] + ([("Q_in by vector direction", r[2]), ("Q_out by vector direction", r[3])] if len(r) == 4 else [])   )
         
                

   # ************* data statistics ****************************
   
   def on_calc_data_statistics(self):
      if not self.is_prepared(["mesh","element_data"]): return
      action_name = self.get_action_name_by_myname()
      if (thread:=self.perform_isolated_action(action_name, self.diskmod.get_element_data_references_physicals_names)) is None: return
      d = DataStatisticsDialog(self, action_name, self.diskmod.is_DEM_loaded())
      d.set_inputs(thread.result[1], thread.result[0])      
      if d.execute_modal():
         self.ex_calc_data_statistics((self.gen_action_id(), action_name, d.out_wfa_ids),
           d.get_protocol_inputs(),
           physicals_ids=d.out_sel_physicals_ids, **d.out_range,
           data_refs=d.out_sel_data_refs, weights_type=d.out_weights_type, quantiles=d.out_quantiles, hist_bins=d.out_histogram, hist_as_pdf=d.out_histogram_PDF,
           file_name_rows=d.out_file_name_rows, file_name_cols=d.out_file_name_cols, include_data_file_name=False, to_protocol=d.out_to_protocol, col_sep=d.out_col_sep, dec_comma=d.out_dec_comma, num_format=d.out_num_format)            
   

   def ex_calc_data_statistics(self, id_name_wait, protocol_inputs, **input_params):
      f = self.create_action_frame(*id_name_wait, self.diskmod.calc_e_data_statistics,
           **input_params)
      f.start_with_default_protocoling(
         protocol_inputs,
         lambda r : r )


   # ********* scalars from vectors *********************

   def on_create_scalar_data_from_vectors(self):   
      if not self.is_prepared(["mesh","element_data"]): return
      action_name = self.get_action_name_by_myname()
      if (thread:=self.perform_isolated_action(action_name, self.diskmod.get_vector_element_data_references)) is None: return
      if len(thread.result) == 0:
         msg_info(self, "No vector element data are loaded.")
         return
      d = DataVectorstoScalarsDialog(self, action_name)
      d.set_inputs(thread.result, [])      
      if d.execute_modal():
         self.ex_create_scalar_data_from_vectors((self.gen_action_id(), action_name, d.out_wfa_ids),
             vector_e_data_refs=d.out_selected_refs,                                    
             x_part=d.out_part_x, y_part=d.out_part_y, z_part=d.out_part_z,
             size=d.out_vsize, vert_angle=d.out_vert_angle, horiz_angle=d.out_horiz_angle)
            
   def re_create_scalar_data_from_vectors(self, id_name_wait, rec):
      p = rec[self.pk_inputs]
      r = self.e_data_refs_from_protocol(p)
      q = {k:(n in p["Vectors quantities"]) for k, n in zip(["x_part", "y_part", "z_part", "size", "vert_angle", "horiz_angle"], ["part x", "part y", "part z", "size", "vertical angle", "horizontal angle"])}
      self.ex_create_scalar_data_from_vectors(id_name_wait, vector_e_data_refs=r, **q)

   def ex_create_scalar_data_from_vectors(self, id_name_wait, **input_kwargs):
      f = self.create_action_frame(*id_name_wait, self.diskmod.create_scalar_e_data_from_3D_vectors,
           **input_kwargs)            
      p = self.get_protocol_e_data_refs(input_kwargs["vector_e_data_refs"])
      p.append(("Vectors quantities", [n for k, n in zip(["x_part", "y_part", "z_part", "size", "vert_angle", "horiz_angle"], ["part x", "part y", "part z", "size", "vertical angle", "horizontal angle"]) if input_kwargs[k]]))
      f.start_with_default_protocoling(
         p,
         None )
      

   # ********** calculate data by formula *****************

   def on_calc_data_by_formula(self):   
      if not self.is_prepared(["mesh","element_data"]): return
      action_name = self.get_action_name_by_myname()
      if (thread:=self.perform_isolated_action(action_name, self.diskmod.get_element_data_references)) is None: return
      d = DataByFormulaDialog(self, action_name)
      d.set_inputs(thread.result, [])      
      if d.execute_modal():
         self.ex_calc_data_by_formula((self.gen_action_id(), action_name, d.out_wfa_ids),
             d.out_selected_refs, d.out_new_name, d.out_formula)
            
   def re_calc_data_by_formula(self, id_name_wait, rec):
      p = rec[self.pk_inputs]
      r, n, f = self.e_data_refs_from_protocol(p), p["New quantity name"], p["Formula"]
      self.ex_calc_data_by_formula(id_name_wait, r, n, f)

   def ex_calc_data_by_formula(self, id_name_wait, *args):
      f = self.create_action_frame(*id_name_wait, self.diskmod.calc_new_e_data_by_formula,
           *args)            
      p = self.get_protocol_e_data_refs(args[0])
      f.start_with_default_protocoling(
         p + [("New quantity name",args[1]), ("Formula", "'"+args[2]+"'")],
         lambda new_refs : [("New elements data","")] + [("  - ", "{{Quantity: {:s}, Time: '{:s}', Step: {:n}, File: {:s}}}".format(*a[1:], a[0])) for a in new_refs])


   # ********* depth , Z level raster ************************
   
   def create_protocol_inputs_of_horizontal_raster(self, level, depth, file_name, kwargs):
      w = kwargs
      p = self.get_protocol_e_data_refs(w["data_refs"])
      if level is None:
         p += [("Depth", depth), ("Depth relative to", "DEM" if w["depth_by_DEM"] else "top boudary entity")]
      else: p += [("Z level", level)]      
      xy = [(n, w[v]) for n, v in [("Minimum x","x_min"), ("Maximum x","x_max"), ("Minimum y","y_min"), ("Maximum y","y_max")] if w[v] is not None]   
      if len(xy) > 0:
         p += xy + [("XY position style", "offset to reference position" if w["xy_by_ref"] else "absolute position")]            
      p += [("Cell size",w["cell_size"]), ("Cell center", w["cell_center"])]      
      if len(c:=w["nbs_by_nodes_num"]) > 0:
         p += [("Interpolation parameters", ""),             
               ("  IDW exponent",w["IDW_exp"]),
               ("  Adjacent elements sharing", [str(n)+(" nodes" if n > 1 else " node") for n in c] )]
      else: p += [("Interpolation parameters", "value given by element")]
      if file_name.upper().endswith(".ASC"):
         p.append(("No value", w["value_instead_of_nan"]))
      p.append(("Number format", "'" + w["num_format"] + "'"))   
      p.append(("File name", file_name))
      return p
      

   # ********* data z level raster ****************************
   
   def on_get_z_level_raster(self):
      if not self.is_prepared(["mesh","element_data"]): return
      action_name = self.get_action_name_by_myname()
      if (thread:=self.perform_isolated_action(action_name, self.diskmod.get_element_data_references)) is None: return
      d = DataDepthZlevelRasterDialog (self, action_name, False, False)
      d.set_inputs(thread.result, [])      
      if d.execute_modal():
         self.ex_get_z_level_raster((self.gen_action_id(), action_name, d.out_wfa_ids),
           d.out_level, d.out_file_name,
           data_refs=d.out_selected_refs,
           x_min=d.out_x_min, x_max=d.out_x_max, y_min=d.out_y_min, y_max=d.out_y_max, xy_by_ref=d.out_xy_by_ref, cell_size=d.out_grid_step, cell_center=d.out_cell_center, 
           nbs_by_nodes_num=d.out_nbs_nodes_count, IDW_exp=d.out_IDW_power, value_instead_of_nan=d.out_no_value, num_format=d.out_number_format)


   def ex_get_z_level_raster(self, id_name_wait, *input_args, **input_kwargs):
      f = self.create_action_frame(*id_name_wait, self.diskmod.save_z_level_raster_data,
           *input_args, **input_kwargs)      
      f.start_with_default_protocoling(
         self.create_protocol_inputs_of_horizontal_raster(input_args[0], None, input_args[1], input_kwargs),
         None )


   # ******** data depth level ********************

   def on_get_depth_raster(self):
      if not self.is_prepared(["mesh","element_data"]): return
      action_name = self.get_action_name_by_myname()
      if (thread:=self.perform_isolated_action(action_name, self.diskmod.get_element_data_references)) is None: return
      d = DataDepthZlevelRasterDialog (self, action_name, True, self.diskmod.is_DEM_loaded())
      d.set_inputs(thread.result, [])      
      if d.execute_modal():
         self.ex_get_depth_raster((self.gen_action_id(), action_name, d.out_wfa_ids),
           d.out_level, d.out_file_name,
           depth_by_DEM=d.out_depth_by_DEM, data_refs=d.out_selected_refs,
           x_min=d.out_x_min, x_max=d.out_x_max, y_min=d.out_y_min, y_max=d.out_y_max, xy_by_ref=d.out_xy_by_ref, cell_size=d.out_grid_step, cell_center=d.out_cell_center, 
           nbs_by_nodes_num=d.out_nbs_nodes_count, IDW_exp=d.out_IDW_power, value_instead_of_nan=d.out_no_value, num_format=d.out_number_format)


   def ex_get_depth_raster(self, id_name_wait, *input_args, **input_kwargs):
      f = self.create_action_frame(*id_name_wait, self.diskmod.save_depth_raster_data,
           *input_args, **input_kwargs)      
      f.start_with_default_protocoling(
         self.create_protocol_inputs_of_horizontal_raster(None, *input_args, input_kwargs),
         None )
   

   # ********* [x,y,z], [x,y,depth] points ************************
   
   def create_protocol_inputs_of_points(self, points, points_xy, points_depth, points_names, file_name, kwargs):
      w = kwargs
      p = self.get_protocol_e_data_refs(w["data_refs"])
      if points is None: p += [("Depth relative to", "DEM" if w["depth_by_DEM"] else "top boudary entity")]          
      p += [("XY position style", "offset to reference position" if w["xy_by_ref"] else "absolute position")]            
      p += [("Points", "")]
      if points is None:
         for n, xy, d in zip(points_names, points_xy, points_depth):
            p += [("- Name", n), ("  Position", xy), ("  Depth", d)]
      else:
         for n, xyz in zip(points_names, points):
            p += [("- Name", n), ("  Point", xyz)]
      if len(c:=w["nbs_by_nodes_num"]) > 0:
         p += [("Interpolation parameters", ""),             
               ("  IDW exponent",w["IDW_exp"]),
               ("  Adjacent elements sharing", [str(n)+(" nodes" if len(c)>1 else "node") for n in c] )]
      else: p += [("Interpolation parameters", "value given by element")]
      p.append(("Number format", "'" + w["num_format"] + "'"))   
      p.append(("File name", file_name))
      return p

   
  # *************** data points ****************************
   
   def on_get_data_points(self):
      if not self.is_prepared(["mesh","element_data"]): return
      action_name = self.get_action_name_by_myname()
      if (thread:=self.perform_isolated_action(action_name, self.diskmod.get_element_data_references)) is None: return
      d = DataPointsDialog(self, action_name, False, False)
      d.set_inputs(thread.result, [])      
      if d.execute_modal():
         self.ex_get_data_points((self.gen_action_id(), action_name, d.out_wfa_ids),
           d.out_points, d.out_points_names, d.out_file_name,
           data_refs=d.out_selected_refs, xy_by_ref=d.out_xy_by_ref,
           nbs_by_nodes_num=d.out_nbs_nodes_count, IDW_exp=d.out_IDW_power,
           ref_file_name_col=False, col_sep=d.out_col_sep, dec_comma=d.out_dec_comma, num_format=d.out_num_format,
           write_to_protocol=d.out_write_to_protocol)
      

   def ex_get_data_points(self, id_name_wait, points, points_names, file_name, **input_kwargs):
      f = self.create_action_frame(*id_name_wait, self.diskmod.save_points_data,
           points, points_names, file_name, **input_kwargs)      
      f.start_with_default_protocoling(
         self.create_protocol_inputs_of_points(points, None, None, points_names, file_name, input_kwargs),
         lambda r: r )


  # ********* data points with depth *************************
   
   def on_get_data_points_with_depth(self):
      if not self.is_prepared(["mesh","element_data"]): return
      action_name = self.get_action_name_by_myname()
      if (thread:=self.perform_isolated_action(action_name, self.diskmod.get_element_data_references)) is None: return
      d = DataPointsDialog(self, action_name, True, self.diskmod.is_DEM_loaded())
      d.set_inputs(thread.result, [])      
      if d.execute_modal():
         self.ex_get_data_points_with_depth((self.gen_action_id(), action_name, d.out_wfa_ids),
           d.out_points_xy, d.out_points_depth, d.out_points_names, d.out_file_name,
           data_refs=d.out_selected_refs, xy_by_ref=d.out_xy_by_ref, depth_by_DEM=d.out_depth_by_DEM,
           nbs_by_nodes_num=d.out_nbs_nodes_count, IDW_exp=d.out_IDW_power,
           ref_file_name_col=False, col_sep=d.out_col_sep, dec_comma=d.out_dec_comma, num_format=d.out_num_format,
           write_to_protocol=d.out_write_to_protocol)
      

   def ex_get_data_points_with_depth(self, id_name_wait, points_xy, points_depth, points_names, file_name, **input_kwargs):
      f = self.create_action_frame(*id_name_wait, self.diskmod.save_points_data_with_depth,
           points_xy, points_depth, points_names, file_name, **input_kwargs)      
      f.start_with_default_protocoling(
         self.create_protocol_inputs_of_points(None, points_xy, points_depth, points_names, file_name),
         lambda r: r )



   # **************** show methods *******************


   def get_protocol_selected_file_name(self):
      if len(self.protocol.tag_ranges(tk.SEL)) > 0:      
         s = self.protocol.selection_get().strip()
         s = s.replace("/", "\\")
         if not os.path.isfile(s) or not os.path.exists(s):
            msg_error(self, "Selection '" + s + "'\n" + "is not valid path or file not exists.")
            return
      else:
         msg_error(self, "No path file is selected in protocol.")
         return
      return s


   def on_show_by_gmsh(self):             
      if (s := self.get_protocol_selected_file_name()) is None: return
      p = sp.Popen(["python", self.extent_file_name("mesh_show_process.py")], stdin=sp.PIPE, bufsize=1, text=True, universal_newlines=True)
      p.stdin.write(s+"\n")
      
   
   def on_show_by_notepad(self):
      if (s := self.get_protocol_selected_file_name()) is None: return
      sp.Popen(["notepad", s])


   def on_show_by_msexcel(self):
      if (s := self.get_protocol_selected_file_name()) is None: return
      a = self.options["applications"]["Microsoft Excel"]["executable path"]
      sp.Popen([a, s])      
   

   def on_show_by_librecalc(self):
      if (s := self.get_protocol_selected_file_name()) is None: return
      a = self.options["applications"]["Libre Calc"]["executable path"]
      sp.Popen([a, s])


   def on_show_by_default(self):
      if (s := self.get_protocol_selected_file_name()) is None: return
      os.startfile(s, cwd=os.getcwd())
      

   # ************* help *****************************

   def on_help_documentation(self):
      s = self.extent_file_name("DISKMODtools.pdf")
      os.startfile(s)


   def on_help_about(self):
      s = "DISKMODtools v1.0\n\n\nThe software was developed under the GNU General Public License v3.0."
      msg_info(self, title="About program", message=s)


