import datetime
from time import sleep
import traceback
import threading as thd
import queue

from .messagebox import msg_yesno
from . import theme

# ************** constants *************************

ACTION_ID_ISOLATED = 999999999

PREPARED = "prepared"
WAITING = "waiting"
RUNNING = "running"
COMPLETED = "completed"
TERMINATED = "terminated"


# ************** Info object ************************

class User_termination_error(Exception): pass

class Info:
   
   def __init__(self, tk_master):
      self.ind = {}
      self.start_times = {}
      self.last_timespan = {}
      self.progress_state = {}
      self.queue = queue.Queue(1)
      self.tk_master = tk_master
      

   def allocate_for_current_thread(self):
      r = thd.current_thread()
      self.ind[r] = 0
      self.start_times[r] = []
      self.last_timespan[r] = None
      self.progress_state[r] = None


   def _write(self, s):
      r = thd.current_thread()
      self.queue.put((r.frame.text, s))
      self.tk_master.event_generate("<<action_thread_info>>", when="tail")
      

   def _gettime(self, start_time):
      s = str(datetime.datetime.now() - start_time).split(".")
      if len(s) == 1: return s[0]
      else: return s[0] + "." + s[1][:4]


   def start(self, s, reset=False):
      r = thd.current_thread()
      if reset:
         self.ind[r] = 0
         self.start_times[r].clear()
      self._write(self.ind[r]*".  " + ">>> " + s + ' ...\n')
      self.start_times[r].append(datetime.datetime.now())
      self.ind[r] += 1
      if r.user_termination: raise User_termination_error()


   def __call__(self, s):
      r = thd.current_thread()
      self._write(self.ind[r] * ".  " + s + "\n")
      if r.user_termination: raise User_termination_error()


   def warning(self, s):
      r = thd.current_thread()
      self._write(self.ind[r] * ".  " + "Warning: " + s + "\n")
      if r.user_termination: raise User_termination_error()


   def OK(self, s=None):
      r = thd.current_thread()
      self.ind[r] -= 1
      time = self._gettime(self.start_times[r].pop())
      s = " Result: "+s if s is not None else ""
      self._write(self.ind[r]*".  " + "OK. Duration time: " + time + s + "\n")
      self.last_timespan[r] = time
      if r.user_termination: raise User_termination_error()


   def progress_start(self, s, max_value):
      r = thd.current_thread()
      self.progress_state[r] = [0, max_value]
      self._write(self.ind[r] * ".  " + s + "   0 %")
      if r.user_termination: raise User_termination_error()


   def progress(self, value):
      p = self.progress_state[r:=thd.current_thread()]
      if (a := (100*value) // p[1]) != p[0]:
         p[0] = a
         self._write("\b\b\b\b\b{:3n} %".format(a))
         if a == 100: self._write("\n")
      if r.user_termination: raise User_termination_error()


   def error(self):
      r = thd.current_thread()
      if self.ind[r] > 0:
         self._write("\n" + self.ind[r]*".  " + self._gettime(self.start_times[r][-1]) + " ERROR !\n")
      else: 
         self._write("\n ERROR !\n")
      self._write(traceback.format_exc() + "\n")
      self.ind[r] = 0
      self.start_times[r].clear()


   def process_user_termination(self):
      if thd.current_thread().user_termination:
         raise User_termination_error()


   def free_for_current_thread(self):
      r = thd.current_thread()
      self.ind.pop(r)
      self.start_times.pop(r)
      self.last_timespan.pop(r)
      self.progress_state.pop(r)



# *********** action thread **********************

class ActionThread(thd.Thread):

   def __init__(self, tk_master, frame, action_id, action_name, wait_for_actions_ids, func, *args, **kwargs):
      super().__init__()
      self.tk_master = tk_master
      self.frame = frame
      self.action_id = action_id
      self.action_name = action_name   
      self.wait_for_actions_ids = wait_for_actions_ids
      self.user_termination = False
      self.func = func
      self.args = args
      self.kwargs = kwargs
      self.state = PREPARED


   def run(self):
      (k:=self.tk_master).info.allocate_for_current_thread()
      k.info._write("Action: " + self.action_name + "\n")
      if self.action_id == ACTION_ID_ISOLATED:
         k.info._write("Isolated action\n")
      else: k.info._write("ID: " + str(self.action_id) + "\n")
      if (w:=self.wait_for_actions_ids) is not None:
         self.state = WAITING
         k.info._write("waiting for actions ids: " + ",".join([str(id) for id in w]) + "\n")    
         w = set(w)
         while len(w - {t.action_id for t in k.actions_threads if t.state == COMPLETED}) > 0:
            sleep(0.2)
      self.start_time = datetime.datetime.now()
      k.info._write("Starting time: " + str(self.start_time).split(".")[0] + "\n")     
      k.event_generate("<<action_thread_start>>", when="tail", state=self.action_id)
      try:
         self.state = RUNNING
         self.result = self.func(*self.args, **self.kwargs)         
      except BaseException as e:
         k.info.error()
         self.terminated_by_exception = e
         self.is_terminated_by_user = e.__class__ is User_termination_error
         self.duration_time = datetime.datetime.now() - self.start_time
         k.info.free_for_current_thread()
         d = ["tk_master","frame","func","args","kwargs"]
         for a in d: delattr(self, a)
         self.state = TERMINATED
         k.event_generate("<<action_thread_terminated>>", when="tail", state=self.action_id)
      else:
         self.duration_time = datetime.datetime.now() - self.start_time
         k.info.free_for_current_thread()
         d = ["frame","func","args","kwargs"]
         for a in d: delattr(self, a)
         self.state = COMPLETED
         k.event_generate("<<action_thread_completed>>", when="tail", state=self.action_id)      
      

   def wait(self):
      while self.is_alive():
         sleep(0.01)
         self.tk_master.update()
      return self.state == COMPLETED


   def send_user_termination(self):
      self.user_termination = True
   
      

# ********** Action Frame ******************************

class ActionFrame(theme.Action_Frame):

      def __init__(self, tk_master, action_id, action_name, wait_for_actions_ids, func, *args, **kwargs):
         super().__init__(tk_master.workarea.actions, width=150, height=100)
         self.columnconfigure(0, weight=0)
         self.columnconfigure(1, weight=1)
         self.rowconfigure(0, weight=0)
         self.rowconfigure(1, weight=1)
         self.label = theme.Action_Label(self, text=action_name)
         self.label.grid(row=0, column=0, sticky="W", padx=5, pady=2)
         self.stop_button = theme.Action_Button(self, text="Stop", command=self.ask_user_terminated)
         self.stop_button.grid(row=0, column=1, sticky="W", padx=2, pady=2)
         self.text = theme.Action_ScrolledText(self, wrap="none")
         self.text.grid(row=1, column=0, columnspan=2, sticky="SNEW")      
         self.log_saved = False
         self.do_before_start = None
         self.do_after_completed = None
         self.do_after_terminated = None
         self.thread = ActionThread(tk_master, self, action_id, action_name, wait_for_actions_ids, func, *args, **kwargs)


      def set_do_before_start(self, func, *args, **kwargs):
         self.do_before_start = func
         self.do_before_start_args = args
         self.do_before_start_kwargs = kwargs


      def set_do_after_completed(self, func, *args, **kwargs):
         self.do_after_completed = func
         self.do_after_completed_args = args
         self.do_after_completed_kwargs = kwargs


      def set_do_after_terminated(self, func, *args, **kwargs):
         self.do_after_terminated = func
         self.do_after_terminated_args = args
         self.do_after_terminated_kwargs = kwargs


      def start_with_default_protocoling(self, inputs, results):
         m = self.thread.tk_master
         self.set_do_before_start(
            m.protocol_start_action, 
            inputs=inputs )
         self.set_do_after_completed(
            m.protocol_complete_action,
            results=results )
         self.set_do_after_terminated(
            m.protocol_terminate_action )
         self.thread.start()


      def start_with_option_protocoling(self, options):
         m = self.thread.tk_master
         self.set_do_after_completed(
            m.protocol_perform_action,
            options=options )
         self.set_do_after_terminated(
            m.protocol_terminate_action )
         self.thread.start()
      

      def ask_user_terminated(self):
#         if msg_yesno(self.thread.tk_master, "Do you really terminate action?"):
            if self.stop_button.winfo_exists():
               self.stop_button["state"] = "disabled"
               self.stop_button["text"] = "Stopping"
               self.thread.send_user_termination()
            




